import io
import re
import unittest.mock
from contextlib import redirect_stdout

import numpy as np
import pandas as pd
import pytest

from seeq import spy
from seeq.spy import _common
from seeq.spy._status import Status
from seeq.spy.assets._trees import Tree
from seeq.spy.assets._trees import _constants, _csv, _match, _path, _properties, _utils, _validate
from seeq.spy.assets._trees.tests.test_tree_system import assert_tree_equals_expected


def assert_frame_equal(df1, df2):
    # noinspection PyProtectedMember
    return pd._testing.assert_frame_equal(df1.sort_index(axis=1),
                                          df2.sort_index(axis=1),
                                          check_dtype=False)


def _tree_from_nested_dict(d):
    if len(d) != 1:
        raise ValueError('Cannot have more than one root.')

    root_name, root_branches = [(k, v) for k, v in d.items()][0]
    tree = Tree(root_name)

    def _add_branches(parent_name, branches_dict):
        for branch_name, sub_branches in branches_dict.items():
            tree.insert(branch_name, parent_name)
            _add_branches(branch_name, sub_branches)

    _add_branches(root_name, root_branches)
    return tree


def _build_dataframe_from_path_name_depth_triples(data):
    df = pd.DataFrame(columns=_constants.dataframe_columns)
    return df.append([{
        'Type': 'Asset',
        'Path': path,
        'Depth': depth,
        'Name': name,
    } for path, name, depth in data])


@pytest.mark.unit
def test_constructor_invalid():
    # Basic property validation
    with pytest.raises(TypeError, match="Argument 'data' should be type DataFrame or str, but is type int"):
        Tree(0)
    with pytest.raises(TypeError, match="'data' must be a name, name of a csv file, Seeq ID, or Metadata dataframe"):
        Tree(data='')
    with pytest.raises(TypeError, match="'data' must be a name, name of a csv file, Seeq ID, or Metadata dataframe"):
        Tree(None)
    with pytest.raises(ValueError, match="DataFrame with no rows"):
        Tree(pd.DataFrame(columns=['Name']))
    with pytest.raises(TypeError, match="Argument 'description' should be type str"):
        Tree('name', description=0)
    with pytest.raises(TypeError, match="Argument 'workbook' should be type str"):
        Tree('name', workbook=0)
    with pytest.raises(TypeError, match="should be type DataFrame or str, but is type Tree"):
        Tree(Tree('Tree Inception'))

    df = pd.DataFrame([{'Name': 'root1', 'Type': 'Asset'}, {'Name': 'root2', 'Type': 'Asset'}])
    with pytest.raises(RuntimeError, match="A tree can only have one root"):
        Tree(df)

    with pytest.raises(RuntimeError, match="Not logged in"):
        Tree('8DEECF16-A500-4231-939D-6C24DD123A30')


@pytest.mark.unit
def test_constructor_name():
    # Valid constructor for a new root asset with all other properties default
    name = 'test name'
    expected = _build_dataframe_from_path_name_depth_triples([
        ('', 'test name', 1)
    ])
    test_tree = Tree(name)
    assert test_tree._dataframe.columns.equals(expected.columns)
    assert test_tree._dataframe.iloc[0].equals(expected.iloc[0])
    assert test_tree._workbook == spy._common.DEFAULT_WORKBOOK_PATH

    # Valid constructor for a new root asset with all other properties assigned to non-defaults
    description = 'test description'
    workbook = 'test workbook'
    expected['Description'] = [description]
    test_tree = Tree(name, description=description, workbook=workbook)
    assert_frame_equal(test_tree._dataframe, expected)
    assert test_tree._workbook == workbook


@pytest.mark.unit
def test_insert_by_name():
    tree_dict = {
        'Root Asset': {
            'L Asset': {
                'LL Asset': {},
                'LR Asset': {}
            },
            'R Asset': {}
        }
    }
    test_tree = _tree_from_nested_dict(tree_dict)
    expected = _build_dataframe_from_path_name_depth_triples([
        ('', 'Root Asset', 1),
        ('Root Asset', 'L Asset', 2),
        ('Root Asset >> L Asset', 'LL Asset', 3),
        ('Root Asset >> L Asset', 'LR Asset', 3),
        ('Root Asset', 'R Asset', 2),
    ])
    assert_frame_equal(test_tree._dataframe, expected)


@pytest.mark.unit
def test_insert_by_name_list():
    tree_dict = {
        'Root Asset': {
            'Location A': {},
            'Location B': {}
        }
    }
    test_tree = _tree_from_nested_dict(tree_dict)
    test_tree.insert([f'Equipment {n}' for n in range(1, 4)], parent='Location A')
    expected = _build_dataframe_from_path_name_depth_triples([
        ('', 'Root Asset', 1),
        ('Root Asset', 'Location A', 2),
        ('Root Asset >> Location A', 'Equipment 1', 3),
        ('Root Asset >> Location A', 'Equipment 2', 3),
        ('Root Asset >> Location A', 'Equipment 3', 3),
        ('Root Asset', 'Location B', 2),
    ])
    assert_frame_equal(test_tree._dataframe, expected)


@pytest.mark.unit
def test_insert_at_depth():
    tree_dict = {
        'Root Asset': {
            'Location A': {},
            'Location B': {}
        }
    }
    test_tree = _tree_from_nested_dict(tree_dict)
    test_tree.insert([f'Equipment {n}' for n in range(1, 4)], parent=2)
    expected = _build_dataframe_from_path_name_depth_triples([
        ('', 'Root Asset', 1),
        ('Root Asset', 'Location A', 2),
        ('Root Asset >> Location A', 'Equipment 1', 3),
        ('Root Asset >> Location A', 'Equipment 2', 3),
        ('Root Asset >> Location A', 'Equipment 3', 3),
        ('Root Asset', 'Location B', 2),
        ('Root Asset >> Location B', 'Equipment 1', 3),
        ('Root Asset >> Location B', 'Equipment 2', 3),
        ('Root Asset >> Location B', 'Equipment 3', 3),
    ])
    assert_frame_equal(test_tree._dataframe, expected)


@pytest.mark.unit
def test_insert_at_path():
    tree_dict = {
        'Root Asset': {
            'Factory': {
                'Location A': {},
                'Location B': {}
            }
        }
    }
    test_tree = _tree_from_nested_dict(tree_dict)
    # Test partial path match with regex
    test_tree.insert('Equipment 1', parent='Factory >> Location [A-Z]')
    # Test full path match with case insensitivity
    test_tree.insert('Equipment 2', parent='rOoT aSsEt >> FaCtOrY >> lOcAtIoN b')
    expected = _build_dataframe_from_path_name_depth_triples([
        ('', 'Root Asset', 1),
        ('Root Asset', 'Factory', 2),
        ('Root Asset >> Factory', 'Location A', 3),
        ('Root Asset >> Factory >> Location A', 'Equipment 1', 4),
        ('Root Asset >> Factory', 'Location B', 3),
        ('Root Asset >> Factory >> Location B', 'Equipment 1', 4),
        ('Root Asset >> Factory >> Location B', 'Equipment 2', 4),
    ])
    assert_frame_equal(test_tree._dataframe, expected)


@pytest.mark.unit
def test_insert_at_root():
    tree_dict = {
        'Root Asset': {
            'Location A': {},
            'Location B': {}
        }
    }
    test_tree = _tree_from_nested_dict(tree_dict)
    test_tree.insert('Location C')
    expected = _build_dataframe_from_path_name_depth_triples([
        ('', 'Root Asset', 1),
        ('Root Asset', 'Location A', 2),
        ('Root Asset', 'Location B', 2),
        ('Root Asset', 'Location C', 2),
    ])
    assert_frame_equal(test_tree._dataframe, expected)


@pytest.mark.unit
def test_insert_at_regex():
    tree_dict = {
        'Root Asset': {
            'Factory': {
                'Location Z': {}
            },
            'Area 51': {}
        }
    }
    test_tree = _tree_from_nested_dict(tree_dict)
    test_tree.insert('Equipment 1', parent='Area [1-9][0-9]*|Location [A-Z]+')
    expected = _build_dataframe_from_path_name_depth_triples([
        ('', 'Root Asset', 1),
        ('Root Asset', 'Area 51', 2),
        ('Root Asset >> Area 51', 'Equipment 1', 3),
        ('Root Asset', 'Factory', 2),
        ('Root Asset >> Factory', 'Location Z', 3),
        ('Root Asset >> Factory >> Location Z', 'Equipment 1', 4)
    ])
    assert_frame_equal(test_tree._dataframe, expected)


@pytest.mark.unit
def test_insert_at_glob():
    tree_dict = {
        'Root Asset': {
            'Location A': {},
            'Location 1': {}
        }
    }
    test_tree = _tree_from_nested_dict(tree_dict)
    test_tree.insert('Equipment 1', parent='Location ?')
    expected = _build_dataframe_from_path_name_depth_triples([
        ('', 'Root Asset', 1),
        ('Root Asset', 'Location 1', 2),
        ('Root Asset >> Location 1', 'Equipment 1', 3),
        ('Root Asset', 'Location A', 2),
        ('Root Asset >> Location A', 'Equipment 1', 3)
    ])
    assert_frame_equal(test_tree._dataframe, expected)


@pytest.mark.unit
def test_insert_preexisting_node():
    tree_dict = {
        'Root': {
            'Location A': {}
        }
    }
    tree = _tree_from_nested_dict(tree_dict)
    tree.insert('lOcAtIoN a')
    expected = _tree_from_nested_dict(tree_dict)
    assert_frame_equal(tree._dataframe, expected._dataframe)


@pytest.mark.unit
def test_insert_same_node_twice():
    tree_dict = {
        'Root': {}
    }
    tree = _tree_from_nested_dict(tree_dict)
    expected_dict = {
        'Root': {
            'Location A': {}
        }
    }
    expected = _tree_from_nested_dict(expected_dict)
    tree.insert(['Location A', 'Location A'])
    assert_frame_equal(tree._dataframe, expected._dataframe)


@pytest.mark.unit
def test_insert_no_parent_match():
    tree = Tree('Root')

    status = Status()
    tree.insert(children=['Child 1', 'Child 2'], parent=3, status=status)
    assert 'No matching parents found. Nothing was inserted.' in status.warnings

    status = Status()
    tree.insert(children=['Child 1', 'Child 2'], parent='asdf', status=status)
    assert 'No matching parents found. Nothing was inserted.' in status.warnings


@pytest.mark.unit
def test_constructor_dataframe_implied_and_leading_assets():
    # The constructor will imply assets and remove redundant leading assets.
    # Even though 'Root' and 'Location B' are not explicitly stated, they must exist for this to be a valid tree.
    insertions = _build_dataframe_from_path_name_depth_triples([
        ('Redundant >> Assets >> Will >> Be >> Removed >> Root', 'Location A', 7),
        ('Redundant >> Assets >> Will >> Be >> Removed >> Root >> Location A', 'Equipment 1', 8),
        ('Redundant >> Assets >> Will >> Be >> Removed >> Root >> Location B', 'Equipment 2', 8),
    ])
    tree = Tree(insertions)
    expected = _build_dataframe_from_path_name_depth_triples([
        ('', 'Root', 1),
        ('Root', 'Location A', 2),
        ('Root >> Location A', 'Equipment 1', 3),
        ('Root', 'Location B', 2),
        ('Root >> Location B', 'Equipment 2', 3),
    ])
    assert_frame_equal(tree._dataframe, expected)

    # And try with Path+Asset columns
    insertions = _build_dataframe_from_path_name_depth_triples([
        ('Redundant >> Assets >> Will >> Be >> Removed >> Root', 'Equipment 1', 8),
        ('Redundant >> Assets >> Will >> Be >> Removed >> Root', 'Equipment 2', 8),
    ])
    insertions['Asset'] = ['Location A', 'Location B']
    tree = Tree(insertions)
    assert_frame_equal(tree._dataframe, expected)


@pytest.mark.unit
def test_insert_dataframe_implied_and_leading_assets():
    expected = _build_dataframe_from_path_name_depth_triples([
        ('', 'Root', 1),
        ('Root', 'Location A', 2),
        ('Root >> Location A', 'Equipment 1', 3),
        ('Root', 'Location B', 2),
        ('Root >> Location B', 'Equipment 2', 3),
    ])
    insertions = _build_dataframe_from_path_name_depth_triples([
        ('Redundant >> Assets >> Will >> Be >> Removed >> Root >> Location A', 'Equipment 1', 8),
        ('Redundant >> Assets >> Will >> Be >> Removed >> Root >> Location B', 'Equipment 2', 8),
    ])
    tree = Tree('Root')
    status = Status()
    tree.insert(insertions, status=status)
    assert_frame_equal(tree._dataframe, expected)
    assert status.df.squeeze()['Total Items Inserted'] == 4

    # And try with Path+Asset columns
    insertions = _build_dataframe_from_path_name_depth_triples([
        ('Redundant >> Assets >> Will >> Be >> Removed >> Root', 'Equipment 1', 8),
        ('Redundant >> Assets >> Will >> Be >> Removed >> Root', 'Equipment 2', 8),
    ])
    insertions['Asset'] = ['Location A', 'Location B']
    tree = Tree('Root')
    status = Status()
    tree.insert(insertions, status=status)
    assert_frame_equal(tree._dataframe, expected)
    assert status.df.squeeze()['Total Items Inserted'] == 4


@pytest.mark.unit
def test_insert_dataframe_name_only():
    expected1 = _build_dataframe_from_path_name_depth_triples([
        ('', 'Root', 1),
        ('Root', 'Location A', 2),
        ('Root', 'Location B', 2),
    ])
    insertions1 = pd.DataFrame([{'Name': 'Location A'}, {'Name': 'Location B'}])
    tree = Tree('Root')
    status = Status()
    tree.insert(insertions1, status=status)
    assert_frame_equal(tree._dataframe, expected1)
    assert status.df.squeeze()['Total Items Inserted'] == 2
    assert status.df.squeeze()['Assets Inserted'] == 2

    expected2 = _build_dataframe_from_path_name_depth_triples([
        ('', 'Root', 1),
        ('Root', 'Location A', 2),
        ('Root >> Location A', 'Equipment 1', 3),
        ('Root >> Location A', 'Equipment 2', 3),
        ('Root', 'Location B', 2),
        ('Root >> Location B', 'Equipment 1', 3),
        ('Root >> Location B', 'Equipment 2', 3),
    ])
    insertions2 = pd.DataFrame([{'Name': 'Equipment 1'}, {'Name': 'Equipment 2'}])
    status = Status()
    tree.insert(insertions2, parent='location *', status=status)
    assert_frame_equal(tree._dataframe, expected2)
    assert status.df.squeeze()['Total Items Inserted'] == 4
    assert status.df.squeeze()['Assets Inserted'] == 4


@pytest.mark.unit
def test_insert_dataframe_missing_name():
    insertions = pd.DataFrame([{'Formula': 'days()'}])
    tree = Tree('Root')
    with pytest.raises(RuntimeError, match="'Name' or 'Friendly Name' is required"):
        tree.insert(insertions)


@pytest.mark.unit
def test_insert_non_string_children_with_formula_or_formula_params_dataframe_case():
    tree = Tree('Children Tree')
    children_df = pd.DataFrame([{'Name': 'Child 1'}, {'Name': 'Child 2'}])
    tree.insert(children_df, formula='sinusoid()')
    expected = list()
    expected.append({
        'Name': 'Children Tree',
        'Path': '',
        'Type': 'Asset'
    })
    expected.append({
        'Name': 'Child 1',
        'Path': 'Children Tree',
        'Formula': 'sinusoid()'
    })
    expected.append({
        'Name': 'Child 2',
        'Path': 'Children Tree',
        'Formula': 'sinusoid()'
    })
    assert_tree_equals_expected(tree, expected)

    # second case (testing with formula parameters)
    children_df_2 = pd.DataFrame([{'Name': 'Child 3'}, {'Name': 'Child 4'}])
    tree.insert(children_df_2, formula='$x', formula_parameters='$x=Child 1', parent='Children Tree')
    expected.append({
        'Name': 'Child 3',
        'Path': 'Children Tree',
        'Formula': '$x',
        'Formula Parameters': {'$x': 'Child 1'}
    })
    expected.append({
        'Name': 'Child 4',
        'Path': 'Children Tree',
        'Formula': '$x',
        'Formula Parameters': {'$x': 'Child 1'}
    })
    assert_tree_equals_expected(tree, expected)


@pytest.mark.unit
def test_insert_tree_and_formula_case_fail():
    tree = Tree('Main Tree')
    child_tree = Tree('Child Tree')
    children_df = pd.DataFrame([{'Name': 'Child 1'}, {'Name': 'Child 2'}])
    child_tree.insert(children_df)
    with pytest.raises(Exception, match="Children DataFrame cannot contain a 'Formula' or 'Formula Parameters'"):
        tree.insert(child_tree, formula='sinusoid()')
    with pytest.raises(Exception, match="Children DataFrame cannot contain a 'Formula' or 'Formula Parameters'"):
        tree.insert(child_tree, formula_parameters='$x=sinusoid()', formula='$x')


@pytest.mark.unit
def test_dataframe_with_assets_insert_with_formula_fail_case():
    tree = Tree('dataframe_case')
    children_df = pd.DataFrame([{'Name': 'Asset 1', 'Type': 'Asset'}])
    with pytest.raises(Exception, match="Assets cannot have formulas or formula parameters."):
        tree.insert(children_df, formula='sinusoid')


@pytest.mark.unit
def test_children_used_as_friendly_name_string_case():
    tree = Tree('children_as_name')
    tree.insert('My Calculation', formula='some_formula()')
    expected = list()
    expected.append({
        'Name': 'children_as_name',
        'Path': '',
        'Type': 'Asset'
    })
    expected.append({
        'Name': 'My Calculation',
        'Path': 'children_as_name',
        'Type': pd.NA,
        'Formula': 'some_formula()'
    })
    assert_tree_equals_expected(tree, expected)

    # second case (testing with formula parameters)
    param_id = _common.new_placeholder_guid()
    tree.insert('My Second Calculation', formula='$x', formula_parameters=f'$x={param_id}',
                parent='children_as_name')
    expected.append({
        'Name': 'My Second Calculation',
        'Path': 'children_as_name',
        'Type': pd.NA,
        'Formula': '$x',
        'Formula Parameters': {'$x': param_id}
    })
    assert_tree_equals_expected(tree, expected)


@pytest.mark.unit
def test_children_used_as_friendly_name_list_case():
    tree = Tree('children_list_as_name')
    children_list = ['My Calculation', 'My Second Calculation']
    tree.insert(children_list, formula='some_formula()')
    expected = list()
    expected.append({
        'Name': 'children_list_as_name',
        'Path': '',
        'Type': 'Asset'
    })
    expected.append({
        'Name': 'My Calculation',
        'Path': 'children_list_as_name',
        'Type': pd.NA,
        'Formula': 'some_formula()'
    })
    expected.append({
        'Name': 'My Second Calculation',
        'Path': 'children_list_as_name',
        'Type': pd.NA,
        'Formula': 'some_formula()'
    })
    assert_tree_equals_expected(tree, expected)

    # second case (testing with formula parameters)
    param_id = _common.new_placeholder_guid()
    children_list_2 = ['My Third Calculation', 'My Fourth Calculation']
    tree.insert(children_list_2, formula='$x', formula_parameters=f'$x={param_id}',
                parent='children_list_as_name')
    expected.append({
        'Name': 'My Third Calculation',
        'Path': 'children_list_as_name',
        'Type': pd.NA,
        'Formula': '$x',
        'Formula Parameters': {'$x': param_id}
    })
    expected.append({
        'Name': 'My Fourth Calculation',
        'Path': 'children_list_as_name',
        'Type': pd.NA,
        'Formula': '$x',
        'Formula Parameters': {'$x': param_id}
    })
    assert_tree_equals_expected(tree, expected)


@pytest.mark.unit
def test_insert_missing_input():
    tree = Tree('test_tree_system10')
    tree.insert('Asset')

    tree.insert(pd.DataFrame([{
        'Name': 'Insert Invalid Formula',
        'Formula': 'sinusoid()'
    }]), parent='Asset')

    with pytest.raises(Exception, match='Must have a Formula if Formula Parameters are defined.'):
        tree.insert(pd.DataFrame([{
            'Name': 'test',
            'Formula Parameters': 'x=Insert Invalid Formula'
        }]), parent='Asset')

    with pytest.raises(Exception, match='All Signals and Conditions require either a formula or an ID'):
        tree.insert(pd.DataFrame([{
            'Name': 'Insert Empty Formula',
            'Type': 'Signal',
            'Formula': ''
        }]), parent='Asset')

    with pytest.raises(Exception, match="The property 'Type' is required for all items without formulas."):
        tree.insert(pd.DataFrame([{
            'Name': 'Insert Invalid Formula No Type',
            'Formula': ''
        }]), parent='Asset')

    with pytest.raises(Exception, match="The property 'Name' or 'Friendly Name' is required"):
        tree.insert(pd.DataFrame([{
            'Formula': 'sinusoid()'
        }]), parent='Asset')


@pytest.mark.unit
def test_formula_params_to_dict_strip():
    tree = Tree('Cases with Spaces')
    param_id = _common.new_placeholder_guid()
    tree.insert(formula='$x', formula_parameters=f'  x    =    {param_id}  ', name='Spacey')
    expected = list()
    expected.append({
        'Name': 'Cases with Spaces',
        'Path': '',
        'Type': 'Asset'
    })
    expected.append({
        'Name': 'Spacey',
        'Path': 'Cases with Spaces',
        'Type': pd.NA,
        'Formula': '$x',
        'Formula Parameters': {'x': param_id}
    })


@pytest.mark.unit
def test_formula_params_to_dict():
    pandas_na_input = _properties.formula_parameters_to_dict(pd.NA)
    assert pd.isnull(pandas_na_input)

    none_input = _properties.formula_parameters_to_dict(None)
    assert pd.isnull(none_input)

    empty_string_input = _properties.formula_parameters_to_dict('')
    assert empty_string_input == {}

    empty_list_input = _properties.formula_parameters_to_dict([])
    assert empty_list_input == {}

    empty_dict_input = _properties.formula_parameters_to_dict({})
    assert empty_dict_input == {}

    string_input = _properties.formula_parameters_to_dict('$x=sinusoid()')
    assert string_input == {'$x': 'sinusoid()'}

    series_data = ['$x=sinusoid()', '$y=cosinusoid()']
    series_index = [1, 2]
    series_input = _properties.formula_parameters_to_dict(pd.Series(data=series_data, index=series_index))
    assert series_input == {'$x': 'sinusoid()', '$y': 'cosinusoid()'}

    list_input = _properties.formula_parameters_to_dict(['$x = sinusoid()', '$y=cosinusoid()'])
    assert list_input == {'$x': 'sinusoid()', '$y': 'cosinusoid()'}


@pytest.mark.unit
def test_formula_insert():
    tree = Tree('Test Formula Insert')

    tree.insert(formula='$x', formula_parameters='x=A1C58783-5951-413F-A584-D73CED2C0191',
                name='Test Formula Parameters')
    tree.insert(friendly_name='Test No Formula Parameters', formula='sinusoid()')
    expected = list()
    expected.append({
        'Name': 'Test Formula Insert',
        'Path': '',
        'Type': 'Asset'
    })
    expected.append({
        'Name': 'Test Formula Parameters',
        'Path': 'Test Formula Insert',
        'Type': pd.NA,
        'Formula Parameters': {'x': 'A1C58783-5951-413F-A584-D73CED2C0191'},
        'Formula': '$x'
    })
    expected.append({
        'Name': 'Test No Formula Parameters',
        'Path': 'Test Formula Insert',
        'Type': pd.NA,
        'Formula': 'sinusoid()'
    })
    assert_tree_equals_expected(tree, expected)

    with pytest.raises(Exception, match='If no `children` argument is given, exactly one of the following'):
        tree.insert(formula='$x')

    with pytest.raises(Exception, match='Only one of the following arguments may be given'):
        tree.insert(children='hello', name='bye', formula='days()')


@pytest.mark.unit
def test_formula_parameters_ignored():
    tree = Tree('Ignore Formula Parameters')
    dataframe = pd.DataFrame([{'Name': 'Asset 1'}, {'Name': 'Asset 2'}])
    tree.insert(dataframe, formula_parameters='$x=ignore()')
    expected = list()
    expected.append({
        'Name': 'Ignore Formula Parameters',
        'Path': '',
        'Type': 'Asset'
    })
    expected.append({
        'Name': 'Asset 1',
        'Path': 'Ignore Formula Parameters',
        'Type': 'Asset'
    })
    expected.append({
        'Name': 'Asset 2',
        'Path': 'Ignore Formula Parameters',
        'Type': 'Asset'
    })
    assert_tree_equals_expected(tree, expected)


@pytest.mark.unit
def test_formula_parameters_to_dict():
    formula_parameters_empty = _properties.formula_parameters_to_dict(None)
    assert pd.isnull(formula_parameters_empty)

    formula_string = 'x=42'
    formula_parameters_string = _properties.formula_parameters_to_dict(formula_string)
    assert isinstance(formula_parameters_string, dict)
    expected_dict_string = {'x': '42'}
    shared_items = {k: formula_parameters_string[k] for k in formula_parameters_string if
                    k in expected_dict_string and formula_parameters_string[k]
                    == expected_dict_string[k]}
    assert len(shared_items) == 1

    formula_list = ['x=1', 'y=2', 'z=3', 'a=4']
    formula_parameters_list = _properties.formula_parameters_to_dict(formula_list)
    assert isinstance(formula_parameters_list, dict)
    expected_dict_list = {'x': '1', 'y': '2', 'z': '3', 'a': '4'}
    shared_items = {k: formula_parameters_list[k] for k in formula_parameters_list if
                    k in expected_dict_list and formula_parameters_list[k]
                    == expected_dict_list[k]}
    assert len(shared_items) == 4

    invalid_formula_syntax_empty = []
    formula_parameters_empty = _properties.formula_parameters_to_dict(invalid_formula_syntax_empty)
    assert isinstance(formula_parameters_empty, dict)
    assert len(formula_parameters_empty) == 0

    invalid_formula_syntax_no_equals = ['x=1', 'y2', 'z=3', 'a=4']
    with pytest.raises(Exception, match='needs to be in the format \'paramName=inputItem\''):
        _properties.formula_parameters_to_dict(invalid_formula_syntax_no_equals)

    invalid_formula_syntax_double_equals = ['x=1', 'y==2', 'z=3', 'a=4']
    with pytest.raises(Exception, match='needs to be in the format \'paramName=inputItem\''):
        _properties.formula_parameters_to_dict(invalid_formula_syntax_double_equals)


@pytest.mark.unit
def test_insert_dataframe_metric():
    insertions = pd.DataFrame([{'Name': 'Location A', 'Type': 'Metric'}])
    tree = Tree('Root')
    status = Status()
    tree.insert(insertions, errors='catalog', status=status)
    assert status.df.squeeze()['Total Items Inserted'] == 0
    assert status.df.squeeze()['Errors Encountered'] == 1


@pytest.mark.unit
def test_insert_dataframe_weird_index():
    expected = _build_dataframe_from_path_name_depth_triples([
        ('', 'Root', 1),
        ('Root', 'Optimizer', 2),
        ('Root', 'Temperature', 2),
    ])
    insertions = pd.DataFrame([{'Name': 'Optimizer'}, {'Name': 'Temperature'}],
                              index=['some index', 'does not actually matter'])
    tree = Tree('Root')
    status = Status()
    tree.insert(insertions, status=status)
    assert_frame_equal(tree._dataframe, expected)
    assert status.df.squeeze()['Total Items Inserted'] == 2


@pytest.mark.unit
def test_insert_dataframe_mixed_scope():
    expected = _build_dataframe_from_path_name_depth_triples([
        ('', 'Root', 1),
        ('Root', 'Optimizer', 2),
        ('Root', 'Temperature', 2),
    ])
    insertions = pd.DataFrame([{'Name': 'Optimizer', 'Scoped To': np.nan},
                               {'Name': 'Temperature', 'Scoped To': '48C3002F-BBEA-4143-8765-D7DADD4E0CA2'}])
    tree = Tree('Root')
    status = Status()
    tree.insert(insertions, status=status)
    assert_frame_equal(tree._dataframe, expected)
    assert status.df.squeeze()['Total Items Inserted'] == 2


@pytest.mark.unit
def test_insert_dataframe_with_mixed_path_existence():
    expected = _build_dataframe_from_path_name_depth_triples([
        ('', 'Root', 1),
        ('Root', 'Location A', 2),
        ('Root >> Location A', 'Equipment 1', 3),
    ])
    # Inserting a NaN path implies that 'Location A' is the sub-root
    insertions = pd.DataFrame([{'Name': 'Location A', 'Path': np.nan},
                               {'Name': 'Equipment 1', 'Path': 'Location A'}])
    tree = Tree('Root')
    status = Status()
    tree.insert(insertions, status=status)
    assert_frame_equal(tree._dataframe, expected)
    assert status.df.squeeze()['Total Items Inserted'] == 2


@pytest.mark.unit
def test_validate_empty_df():
    df = pd.DataFrame(columns=['Any', 'Columns', 'You', 'Want'])
    error_summaries, error_series = _validate.validate(spy.session, df)
    assert len(error_summaries) == 1
    assert 'Tree must be non-empty' in error_summaries[0]
    assert len(error_series) == 0


@pytest.mark.unit
def test_validate_bad_depth():
    df = _build_dataframe_from_path_name_depth_triples([
        ('', 'Root', 1),
        ('Root', 'Location A', 2),
        ('Root >> Location A', 'Equipment 1', 3),
        ('Root >> Location A', 'Equipment 2', 1),
    ])
    error_summaries, error_series = _validate.validate(spy.session, df)
    assert len(error_summaries) == 1
    error_msg = 'Item\'s depth does not match its path'
    assert error_msg in error_summaries[0]
    assert error_msg in error_series[3]

    df = _build_dataframe_from_path_name_depth_triples([
        ('', 'Root', 1),
        ('Root', 'Location A', 3)
    ])
    error_summaries, error_series = _validate.validate(spy.session, df)
    assert len(error_summaries) == 1
    error_msg = 'Item\'s depth does not match its path'
    assert error_msg in error_summaries[0]
    assert error_msg in error_series[1]


@pytest.mark.unit
def test_validate_root():
    df = _build_dataframe_from_path_name_depth_triples([
        ('Super-root', 'Root', 2),
        ('Super-root >> Root', 'Item', 3)
    ])
    error_summaries, error_series = _validate.validate(spy.session, df)
    assert len(error_summaries) == 1
    error_msg = 'The root of the tree cannot be assigned a path.'
    assert error_msg in error_summaries[0]
    assert error_msg in error_series[0]

    df = _build_dataframe_from_path_name_depth_triples([
        ('', 'Root', 1),
        ('Root', 'Item', 2),
        ('', 'Another Root', 1)
    ])
    error_summaries, error_series = _validate.validate(spy.session, df)
    assert len(error_summaries) == 1
    error_msg = 'A tree can only have one root'
    assert error_msg in error_summaries[0]
    assert error_msg in error_series[2]


@pytest.mark.unit
def test_validate_bad_path():
    df = _build_dataframe_from_path_name_depth_triples([
        ('', 'Root', 1),
        ('Root', 'Location A', 2),
        ('Root >> Locat--TYPO--ion A', 'Equipment 1', 3),
        ('Root >> Location A', 'Equipment 2', 3),
    ])
    error_summaries, error_series = _validate.validate(spy.session, df)
    assert len(error_summaries) == 1
    error_msg = 'Item\'s position in tree does not match its path.'
    assert error_msg in error_summaries[0]
    assert error_msg in error_series[2]


@pytest.mark.unit
def test_validate_path_sort():
    df = _build_dataframe_from_path_name_depth_triples([
        ('', 'Root', 1),
        ('Root', 'Location B', 2),
        ('Root >> Location B', 'Equipment 1', 3),
        ('Root', 'Location A', 2),
    ])
    error_summaries, error_series = _validate.validate(spy.session, df)
    assert len(error_summaries) == 1
    error_msg = 'Item is not stored in proper position sorted by path.'
    assert error_msg in error_summaries[0]
    assert error_msg in error_series[3]


@pytest.mark.unit
def test_validate_all_assets_exist():
    df = _build_dataframe_from_path_name_depth_triples([
        ('', 'Root', 1),
        ('Root >> Location A', 'Equipment 1', 3),
        ('Root', 'Location B', 2),
    ])
    error_summaries, error_series = _validate.validate(spy.session, df)
    assert len(error_summaries) == 1
    error_msg = 'Item has an ancestor not stored in this tree.'
    assert error_msg in error_summaries[0]
    assert error_msg in error_series[1]


@pytest.mark.unit
def test_validate_invalid_parent():
    df = _build_dataframe_from_path_name_depth_triples([
        ('', 'Root', 1),
        ('Bad Path', 'Location B', 2),
        ('Bad Path >> Location B', 'Area 1', 3),
        ('Bad Path', 'Location A', 2),
        ('Bad Path >> Location A', 'Area 2', 3),
    ])
    error_summaries, error_series = _validate.validate(spy.session, df)
    assert len(error_summaries) == 2
    assert (error_series.loc[[2, 4]] == 'Item\'s parent is invalid.').all()


@pytest.mark.unit
def test_validate_column_dtypes():
    df = pd.DataFrame([{
        'ID': 3.14159,
        'Referenced ID': -163,
        'Type': pd.Series([1, 2, 3]),
        'Path': set(),
        'Depth': pd.to_datetime('2020'),
        'Name': list(),
        'Description': False,
        'Formula': (),
        'Formula Parameters': 0.577215,
        'Cache Enabled': 'Yes',
        'Aggregation Function': 0.0001,
        'Statistic': [2, 13, 45],
        'Bounding Condition': True,
        'Bounding Condition Maximum Duration': pd.Series([-1, 6]),
        'Duration': 5,
        'Number Format': {"testing": "yes"},
        'Measured Item': [1],
        'Metric Neutral Color': False,
        'Period': 5,
        'Process Type': 80,
        'Thresholds': 1234
    }])
    with unittest.mock.patch('seeq.spy.assets._trees._constants.MAX_ERRORS_DISPLAYED', new=999):
        error_summaries, error_series = _validate.validate(spy.session, df)
    for column in _constants.dataframe_columns:
        error_msg = f"The property '{column}' must have one of the following types"
        assert any([error_msg in x for x in error_summaries]), f"{column} did not cause the expected error"
        assert error_msg in error_series[0], f"{error_msg} was not found in {error_series[0]}"


@pytest.mark.unit
def test_validate_properties():
    df = pd.DataFrame([
        {},
        {'Name': 'My Condition', 'Type': 'Condition'},
        {'Name': 'My Formula Parameters', 'Formula Parameters': '$x=sinusoid()'},
        {'ID': '8DEECF16-A500-4231-939D-6C24DD123A30'},
        {'ID': 'bad-guid-format'},
        {'Referenced ID': 'bad-guid-format'},
        {'Name': 'Area A', 'Path': 'Example >> Cooling Tower 1 >> Cooling Tower 1'},
        {'Name': 'Asset', 'Type': 'Asset', 'Formula': 'sinusoid()'},
    ])
    _, error_series = _validate.validate(spy.session, df, stage='input')
    error_list = [
        "The property 'Name' or 'Friendly Name' is required for all nodes without ID.",
        "Stored Signals and Conditions are not yet supported. All Signals and Conditions require either a formula or "
        "an ID.",
        "Must have a Formula if Formula Parameters are defined.",
        "Must log in via spy.login() before inserting an item via ID or Referenced ID.",
        "The property 'ID' must be a valid GUID. Given: 'bad-guid-format'",
        "The property 'Referenced ID' must be a valid GUID. Given: 'bad-guid-format'",
        "Paths with repeated names are not valid.",
        "Assets cannot have formulas or formula parameters."
    ]

    for i in range(len(error_list)):
        row_validation = _validate.property_validations(spy.session, df.iloc[i], 'input')
        assert row_validation[i][1] in error_series.loc[i]
        assert row_validation[i][1] in error_list[i]

    df = pd.DataFrame([
        {'Name': 'My Asset', 'Path': '', 'Depth': 1},
        {'Name': 'My Condition', 'Type': 'Condition', 'Path': '', 'Depth': 1},
        {'Name': 'My Formula Parameters', 'Type': 'Signal', 'Formula Parameters': '$x=sinusoid()', 'Path': '',
         'Depth': 1},
        {'Path': '', 'Depth': 1, 'Type': 'Asset'},
        {'Name': 'My Asset', 'Depth': 1, 'Type': 'Asset'},
        {'Name': 'My Asset', 'Path': '', 'Type': 'Asset'},
        {'Name': 'Cooling Tower 1', 'Path': 'Example >> Cooling Tower 1', 'Type': 'Asset', 'Depth': 4},
        {'Name': 'Asset', 'Path': '', 'Type': 'Asset', 'Formula': 'sinusoid()', 'Depth': 1},
        {'Name': 'Calculation', 'Path': '', 'Formula': '$x', 'Formula Parameters': '$x=sinusoid()', 'Depth': 1}
    ])
    df['Depth'] = df['Depth'].astype('Int64')
    _, error_series = _validate.validate(spy.session, df, stage='final')
    final_error_list = [
        "The property 'Type' is required for all items without formulas.",
        "Stored Signals and Conditions are not yet supported. All Signals and Conditions require a formula.",
        "Must have a Formula if Formula Parameters are defined.",
        "The property 'Name' is required.",
        "The property 'Path' is required.",
        "The property 'Depth' is required.",
        "Paths with repeated names are not valid.",
        "Assets cannot have formulas or formula parameters."
    ]

    for i in range(len(final_error_list)):
        row_validation = _validate.property_validations(spy.session, df.iloc[i], 'final')
        assert row_validation[i][1] in error_series.loc[i]
        assert row_validation[i][1] in final_error_list[i]


@pytest.mark.unit
def test_validate_metric_properties():
    # Note: Thresholds are tested in the System test
    df = pd.DataFrame([
        {'Name': 'My Metric', 'Type': 'Metric'},
    ])
    _, error_series = _validate.validate(spy.session, df, stage='input')
    assert "Metrics must have a Measured Item or an ID." == error_series.iloc[0]

    df = pd.DataFrame([
        {'Name': 'My Metric', 'Type': 'Metric', 'Measured Item': 'Something', 'Metric Neutral Color': '1234'},
    ])
    _, error_series = _validate.validate(spy.session, df, stage='input')
    assert "Metric neutral color must start with a '#' character and be a valid hex value." == error_series.iloc[0]

    df = pd.DataFrame([
        {'Name': 'My Metric', 'Type': 'Metric', 'Measured Item': 'Something', 'Metric Neutral Color': '#12'},
    ])
    _, error_series = _validate.validate(spy.session, df, stage='input')
    assert "Metric neutral color must start with a '#' character and be a valid hex value." == error_series.iloc[0]

    df = pd.DataFrame([
        {'Name': 'My Metric', 'Type': 'Metric', 'Measured Item': 'Something', 'Statistic': 'Not a real stat'},
    ])
    _, error_series = _validate.validate(spy.session, df, stage='input')
    assert 'Statistic "Not a real stat" not recognized' in error_series.iloc[0]


@pytest.mark.unit
def test_validate_calculations():
    param_guid = _common.new_placeholder_guid()
    df = Tree(pd.DataFrame([{
        'Name': 'Root'
    }, {
        'Name': 'Calculation',
        'Path': 'Root',
        'Formula': '$a + $b',
        'Formula Parameters': {'a': 'Sibling 1', 'b': 'Sibling 2 >> Sibling Child', 'c': param_guid}
    }, {
        'Name': 'Sibling 1',
        'Type': 'Signal',
        'Path': 'Root',
        'Formula': 'whatever'
    }, {
        'Name': 'Sibling 2',
        'Path': 'Root'
    }, {
        'Name': 'Sibling Child',
        'Path': 'Root >> Sibling 2',
        'Type': 'Signal',
        'Formula': 'whatever'
    }]))._dataframe

    # The creation of df tests that formula parameter validation works with good input
    assert df.loc[df.Name == 'Calculation', 'Formula Parameters'].iloc[0] == {'a': 'Sibling 1',
                                                                              'b': 'Sibling 2 >> Sibling Child',
                                                                              'c': param_guid}

    # We now modify df to test bad cases
    msg = 'Formula parameter is invalid, missing, or has been removed from tree: \"'

    df1 = df.copy()
    df1.loc[df1.Name == 'Sibling 1', 'Name'] = 'Modified Sibling 1'
    error_summaries, error_series = _validate.validate(spy.session, df1, stage='final')
    assert len(error_summaries) == 1
    assert (msg + 'Root >> Sibling 1') in error_summaries[0]
    assert (msg + 'Root >> Sibling 1') in error_series[1]

    df2 = df.copy()
    df2.loc[df2.Name == 'Sibling Child', 'Name'] = 'Modified Sibling Child'
    error_summaries, error_series = _validate.validate(spy.session, df2, stage='final')
    assert len(error_summaries) == 1
    assert (msg + 'Root >> Sibling 2 >> Sibling Child') in error_summaries[0]
    assert (msg + 'Root >> Sibling 2 >> Sibling Child') in error_series[1]

    df3 = df.copy()
    df3.loc[df3.Name == 'Sibling 1', 'Name'] = 'Modified Sibling 1'
    df3.loc[df3.Name == 'Sibling Child', 'Name'] = 'Modified Sibling Child'
    error_summaries, error_series = _validate.validate(spy.session, df3, stage='final')
    assert len(error_summaries) == 2
    assert (msg + 'Root >> Sibling 1') in error_summaries[0]
    assert (msg + 'Root >> Sibling 2 >> Sibling Child') in error_summaries[1]
    assert (msg + 'Root >> Sibling 1') in error_series[1]
    assert (msg + 'Root >> Sibling 2 >> Sibling Child') in error_series[1]


@pytest.mark.unit
def test_validate_remove_and_move():
    tree = Tree(pd.DataFrame([{
        'Name': 'Root'
    }, {
        'Name': 'asset',
        'Path': 'Root',
    }, {
        'Name': 'Dependency',
        'Path': 'Root',
        'Type': 'Signal',
        'Formula': 'whatever'
    }, {
        'Name': 'Calculation',
        'Path': 'Root',
        'Type': 'Signal',
        'Formula': '$x + $x',
        'Formula Parameters': {'x': 'Dependency'}
    }]))

    with pytest.raises(RuntimeError, match='Formula parameter is invalid, missing, or has been removed from tree: '
                                           '\"Root >> Dependency\"'):
        tree.remove('Dependency')

    with pytest.raises(RuntimeError, match='Formula parameter is invalid, missing, or has been removed from tree: '
                                           '\"Root >> Dependency\"'):
        tree.move('Dependency', 'asset')


@pytest.mark.unit
def test_insert_other_tree():
    tree_to_insert = Tree('Area A')
    tree_to_insert.insert(['Optimizer', 'Temperature'])

    # Insert a tree directly below the root. The old 'Area A' root will not be transferred over.
    expected_df = _build_dataframe_from_path_name_depth_triples([
        ('', 'Real Root', 1),
        ('Real Root', 'Area A', 2),
        ('Real Root >> Area A', 'Optimizer', 3),
        ('Real Root >> Area A', 'Temperature', 3),
        ('Real Root', 'Tower', 2),
    ])
    tree = Tree('Real Root')
    tree.insert('Tower')
    tree.insert(tree_to_insert)
    assert_frame_equal(tree._dataframe, expected_df)
    # Do it again to show it up-serts the nodes
    tree.insert(tree_to_insert)
    assert_frame_equal(tree._dataframe, expected_df)

    # Insert a tree below multiple parents
    expected_df = _build_dataframe_from_path_name_depth_triples([
        ('', 'Real Root', 1),
        ('Real Root', 'Tower 1', 2),
        ('Real Root >> Tower 1', 'Area A', 3),
        ('Real Root >> Tower 1 >> Area A', 'Optimizer', 4),
        ('Real Root >> Tower 1 >> Area A', 'Temperature', 4),
        ('Real Root', 'Tower 2', 2),
        ('Real Root >> Tower 2', 'Area A', 3),
        ('Real Root >> Tower 2 >> Area A', 'Optimizer', 4),
        ('Real Root >> Tower 2 >> Area A', 'Temperature', 4),
    ])
    tree = Tree('Real Root')
    tree.insert(['Tower 1', 'Tower 2'])
    tree.insert(tree_to_insert, parent='Tower*')
    assert_frame_equal(tree._dataframe, expected_df)


@pytest.mark.unit
def test_trim_unneeded_paths_constructor():
    expected_df = _build_dataframe_from_path_name_depth_triples([
        ('', 'Real Root', 1),
        ('Real Root', 'Tower', 2),
        ('Real Root >> Tower', 'Area A', 3),
        ('Real Root >> Tower >> Area A', 'Optimizer', 4),
        ('Real Root >> Tower >> Area A', 'Temperature', 4),
    ])
    # Test three leading nodes to be removed
    test_df = _build_dataframe_from_path_name_depth_triples([
        ('Dupe Root >> Dupe Path 1 >> Dupe Path 2', 'Real Root', 4),
        ('Dupe Root >> Dupe Path 1 >> Dupe Path 2 >> Real Root', 'Tower', 5),
        ('Dupe Root >> Dupe Path 1 >> Dupe Path 2 >> Real Root >> Tower', 'Area A', 6),
        ('Dupe Root >> Dupe Path 1 >> Dupe Path 2 >> Real Root >> Tower >> Area A', 'Optimizer', 7),
        ('Dupe Root >> Dupe Path 1 >> Dupe Path 2 >> Real Root >> Tower >> Area A', 'Temperature', 7),
    ])
    tree = Tree(test_df)
    assert_frame_equal(tree._dataframe, expected_df)

    # Test one leading node to be removed
    test_df = _build_dataframe_from_path_name_depth_triples([
        ('Dupe Root', 'Real Root', 2),
        ('Dupe Root >> Real Root', 'Tower', 3),
        ('Dupe Root >> Real Root >> Tower', 'Area A', 4),
        ('Dupe Root >> Real Root >> Tower >> Area A', 'Temperature', 5),
        ('Dupe Root >> Real Root >> Tower >> Area A', 'Optimizer', 5),
    ])
    tree = Tree(test_df)
    assert_frame_equal(tree._dataframe, expected_df)

    # Test no changes needed
    test_df = expected_df.copy()
    tree = Tree(test_df)
    assert_frame_equal(tree._dataframe, expected_df)

    # Test with implied shared roots only
    expected_df = _build_dataframe_from_path_name_depth_triples([
        ('', 'Real Root', 1),
        ('Real Root', 'Tower 1', 2),
        ('Real Root >> Tower 1', 'Area A', 3),
        ('Real Root >> Tower 1 >> Area A', 'Optimizer', 4),
        ('Real Root >> Tower 1 >> Area A', 'Temperature', 4),
        ('Real Root', 'Tower 2', 2),
        ('Real Root >> Tower 2', 'Area A', 3),
        ('Real Root >> Tower 2 >> Area A', 'Optimizer', 4),
        ('Real Root >> Tower 2 >> Area A', 'Temperature', 4),
    ])
    test_df = _build_dataframe_from_path_name_depth_triples([
        ('Dupe Path 1 >> Dupe Path 2 >> Real Root >> Tower 1 >> Area A', 'Temperature', 4),
        ('Dupe Path 1 >> Dupe Path 2 >> Real Root >> Tower 1 >> Area A', 'Optimizer', 4),
        ('Dupe Path 1 >> Dupe Path 2 >> Real Root >> Tower 2 >> Area A', 'Temperature', 4),
        ('Dupe Path 1 >> Dupe Path 2 >> Real Root >> Tower 2 >> Area A', 'Optimizer', 4),
    ])
    tree = Tree(test_df)
    assert_frame_equal(tree._dataframe, expected_df)


@pytest.mark.unit
def test_trim_unneeded_paths_insert():
    expected_df = _build_dataframe_from_path_name_depth_triples([
        ('', 'Real Root', 1),
        ('Real Root', 'Tower', 2),
        ('Real Root >> Tower', 'Area A', 3),
        ('Real Root >> Tower >> Area A', 'Optimizer', 4),
        ('Real Root >> Tower >> Area A', 'Temperature', 4),
    ])
    # Test three leading nodes to be removed with the same root as parent
    test_df = _build_dataframe_from_path_name_depth_triples([
        ('Dupe Root >> Dupe Path 1 >> Dupe Path 2', 'Real Root', 4),
        ('Dupe Root >> Dupe Path 1 >> Dupe Path 2 >> Real Root', 'Tower', 5),
        ('Dupe Root >> Dupe Path 1 >> Dupe Path 2 >> Real Root >> Tower', 'Area A', 6),
        ('Dupe Root >> Dupe Path 1 >> Dupe Path 2 >> Real Root >> Tower >> Area A', 'Temperature', 7),
        ('Dupe Root >> Dupe Path 1 >> Dupe Path 2 >> Real Root >> Tower >> Area A', 'Optimizer', 7),
    ])
    tree = Tree('Real Root')
    tree.insert(test_df)
    assert_frame_equal(tree._dataframe, expected_df)

    # Test one leading node to be removed
    test_df = _build_dataframe_from_path_name_depth_triples([
        ('Dupe Root', 'Real Root', 2),
        ('Dupe Root >> Real Root', 'Tower', 3),
        ('Dupe Root >> Real Root >> Tower', 'Area A', 4),
        ('Dupe Root >> Real Root >> Tower >> Area A', 'Temperature', 5),
        ('Dupe Root >> Real Root >> Tower >> Area A', 'Optimizer', 5),
    ])
    tree = Tree('Real Root')
    tree.insert(test_df)
    assert_frame_equal(tree._dataframe, expected_df)

    # Test no changes needed
    tree = Tree('Real Root')
    tree.insert(expected_df.copy())
    assert_frame_equal(tree._dataframe, expected_df)

    # Test three leading nodes to be removed with a different root as parent
    expected_df = _build_dataframe_from_path_name_depth_triples([
        ('', 'Real Root', 1),
        ('Real Root', 'Sub Root', 2),
        ('Real Root >> Sub Root', 'Tower', 3),
        ('Real Root >> Sub Root >> Tower', 'Area A', 4),
        ('Real Root >> Sub Root >> Tower >> Area A', 'Optimizer', 5),
        ('Real Root >> Sub Root >> Tower >> Area A', 'Temperature', 5),
    ])
    test_df = _build_dataframe_from_path_name_depth_triples([
        ('Dupe Root >> Dupe Path 1 >> Dupe Path 2', 'Sub Root', 4),
        ('Dupe Root >> Dupe Path 1 >> Dupe Path 2 >> Sub Root', 'Tower', 5),
        ('Dupe Root >> Dupe Path 1 >> Dupe Path 2 >> Sub Root >> Tower', 'Area A', 6),
        ('Dupe Root >> Dupe Path 1 >> Dupe Path 2 >> Sub Root >> Tower >> Area A', 'Temperature', 7),
        ('Dupe Root >> Dupe Path 1 >> Dupe Path 2 >> Sub Root >> Tower >> Area A', 'Optimizer', 7),
    ])
    tree = Tree('Real Root')
    tree.insert(test_df)
    assert_frame_equal(tree._dataframe, expected_df)

    # Test with implied shared roots only
    expected_df = _build_dataframe_from_path_name_depth_triples([
        ('', 'Real Root', 1),
        ('Real Root', 'Sub Root', 2),
        ('Real Root >> Sub Root', 'Tower 1', 3),
        ('Real Root >> Sub Root >> Tower 1', 'Area A', 4),
        ('Real Root >> Sub Root >> Tower 1 >> Area A', 'Optimizer', 5),
        ('Real Root >> Sub Root >> Tower 1 >> Area A', 'Temperature', 5),
        ('Real Root >> Sub Root', 'Tower 2', 3),
        ('Real Root >> Sub Root >> Tower 2', 'Area A', 4),
        ('Real Root >> Sub Root >> Tower 2 >> Area A', 'Optimizer', 5),
        ('Real Root >> Sub Root >> Tower 2 >> Area A', 'Temperature', 5),
    ])
    test_df = _build_dataframe_from_path_name_depth_triples([
        ('Dupe Path >> Sub Root >> Tower 1 >> Area A', 'Temperature', 5),
        ('Dupe Path >> Sub Root >> Tower 1 >> Area A', 'Optimizer', 5),
        ('Dupe Path >> Sub Root >> Tower 2 >> Area A', 'Temperature', 5),
        ('Dupe Path >> Sub Root >> Tower 2 >> Area A', 'Optimizer', 5),
    ])
    tree = Tree('Real Root')
    tree.insert(test_df)
    assert_frame_equal(tree._dataframe, expected_df)
    # Inserting that same thing again should be idempotent.
    tree.insert(test_df)
    assert_frame_equal(tree._dataframe, expected_df)


@pytest.mark.unit
def test_reify_missing_assets():
    expected_df = _build_dataframe_from_path_name_depth_triples([
        ('', 'Root', 1),
        ('Root', 'Tower', 2),
        ('Root >> Tower', 'Region 1', 3),
        ('Root >> Tower >> Region 1', 'Area A', 4),
        ('Root >> Tower >> Region 1 >> Area A', 'Optimizer', 5),
        ('Root >> Tower >> Region 1 >> Area A', 'Temperature', 5),
        ('Root >> Tower >> Region 1', 'Area B', 4),
        ('Root >> Tower >> Region 1 >> Area B', 'Optimizer', 5),
        ('Root >> Tower', 'Region 2', 3),
        ('Root >> Tower >> Region 2', 'Area C', 4),
        ('Root >> Tower >> Region 2 >> Area C', 'Temperature', 5),
    ])
    # Test everything missing except the leaf nodes
    test_df = _build_dataframe_from_path_name_depth_triples([
        ('Root >> Tower >> Region 1 >> Area A', 'Optimizer', 5),
        ('Root >> Tower >> Region 1 >> Area A', 'Temperature', 5),
        ('Root >> Tower >> Region 1 >> Area B', 'Optimizer', 5),
        ('Root >> Tower >> Region 2 >> Area C', 'Temperature', 5),
    ])
    result_df = _path.reify_missing_assets(test_df)
    _path.sort_by_node_path(result_df)
    assert_frame_equal(result_df, expected_df)

    # Test everything missing between the root and the leaves
    test_df = _build_dataframe_from_path_name_depth_triples([
        ('', 'Root', 1),
        ('Root >> Tower >> Region 1 >> Area A', 'Optimizer', 5),
        ('Root >> Tower >> Region 1 >> Area A', 'Temperature', 5),
        ('Root >> Tower >> Region 1 >> Area B', 'Optimizer', 5),
        ('Root >> Tower >> Region 2 >> Area C', 'Temperature', 5),
    ])
    result_df = _path.reify_missing_assets(test_df)
    _path.sort_by_node_path(result_df)
    assert_frame_equal(result_df, expected_df)

    # Test missing the root-most two levels
    test_df = _build_dataframe_from_path_name_depth_triples([
        ('Root >> Tower', 'Region 1', 3),
        ('Root >> Tower >> Region 1', 'Area A', 4),
        ('Root >> Tower >> Region 1 >> Area A', 'Optimizer', 5),
        ('Root >> Tower >> Region 1 >> Area A', 'Temperature', 5),
        ('Root >> Tower >> Region 1', 'Area B', 4),
        ('Root >> Tower >> Region 1 >> Area B', 'Optimizer', 5),
        ('Root >> Tower', 'Region 2', 3),
        ('Root >> Tower >> Region 2', 'Area C', 4),
        ('Root >> Tower >> Region 2 >> Area C', 'Temperature', 5),
    ])
    result_df = _path.reify_missing_assets(test_df)
    _path.sort_by_node_path(result_df)
    assert_frame_equal(result_df, expected_df)

    # Test no changes needed
    test_df = expected_df.copy()
    result_df = _path.reify_missing_assets(test_df)
    _path.sort_by_node_path(result_df)
    assert_frame_equal(result_df, expected_df)

    # Test where the first two levels should not be reified.
    expected_df = _build_dataframe_from_path_name_depth_triples([
        ('Root >> Tower', 'Region 1', 3),
        ('Root >> Tower >> Region 1', 'Area A', 4),
        ('Root >> Tower >> Region 1 >> Area A', 'Optimizer', 5),
        ('Root >> Tower >> Region 1 >> Area A', 'Temperature', 5),
        ('Root >> Tower >> Region 1', 'Area B', 4),
        ('Root >> Tower >> Region 1 >> Area B', 'Optimizer', 5),
        ('Root >> Tower', 'Region 2', 3),
        ('Root >> Tower >> Region 2', 'Area C', 4),
        ('Root >> Tower >> Region 2 >> Area C', 'Temperature', 5),
    ])
    test_df = _build_dataframe_from_path_name_depth_triples([
        ('Root >> Tower >> Region 1 >> Area A', 'Optimizer', 5),
        ('Root >> Tower >> Region 1 >> Area A', 'Temperature', 5),
        ('Root >> Tower >> Region 1 >> Area B', 'Optimizer', 5),
        ('Root >> Tower >> Region 2 >> Area C', 'Temperature', 5),
    ])
    result_df = _path.reify_missing_assets(test_df, 'Root >> Tower')
    _path.sort_by_node_path(result_df)
    assert_frame_equal(result_df, expected_df)


@pytest.mark.unit
def test_upsert():
    df1 = pd.DataFrame([{
        'Path': 'Root',
        'Name': 'Area A',
        'Property': 'Anything',
        'Numerical': 123
    }, {
        'Path': 'Root >> Area A',
        'Name': 'Temperature',
        'Property': 'Old Value',
        'Numerical': 1,
        'Extra Old Column': 'Anything'
    }])
    df2 = pd.DataFrame([{
        'Path': 'Root >> Area A',
        'Name': 'Optimizer',
        'Property': 'Anything',
        'Numerical': 2,
        'Extra New Column': 'Something Unexpected'
    }, {
        'Path': 'root >> area A',
        'Name': 'temperature',
        'Property': 'New Value',
        'Numerical': np.nan,
        'Extra New Column': 'Something Unexpected'
    }])

    expected_df = pd.DataFrame([{
        'Path': 'Root',
        'Name': 'Area A',
        'Property': 'Anything',
        'Numerical': 123
    }, {
        'Path': 'Root >> Area A',
        'Name': 'Temperature',
        'Property': 'New Value',
        'Numerical': 1,
        'Extra Old Column': 'Anything'
    }, {
        'Path': 'Root >> Area A',
        'Name': 'Optimizer',
        'Property': 'Anything',
        'Numerical': 2
    }])

    upsert_df = _utils.upsert(df1, df2)
    assert_frame_equal(upsert_df, expected_df)
    assert upsert_df['Numerical'].dtype in [np.int32, np.int64]

    expected_df.loc[1, 'Property'] = 'Old Value'
    upsert_df = _utils.upsert(df1, df2, prefer_right=False)
    assert_frame_equal(upsert_df, expected_df)


@pytest.mark.unit
def test_insert_calculation_regex_asset_and_calculation_match():
    workbook = 'insert_several_similar_assets'
    tree = Tree('test_tree_system8', workbook=workbook)
    example_calc_id = _common.new_placeholder_guid()

    tree.insert('Area_A')
    tree.insert('Area_B')
    tree.insert('Area_C')

    tree.insert(pd.DataFrame([{
        'Name': 'Open Door B Please',
        'Formula Parameters': [f'x={example_calc_id}'],
        'Formula': '$x'
    }]), parent='Area_B')

    tree.insert(parent='Area_C', friendly_name='Open Door C Please', formula='$x',
                formula_parameters=[f'x={example_calc_id}'])

    with pytest.raises(RuntimeError, match='matches multiple items in tree'):
        tree.insert(pd.DataFrame([{
            'Name': 'Unspecified Door',
            'Formula Parameters': 'x=Area_* >> open*',
            'Formula': '$x'
        }]), parent='test_tree_system8')


@pytest.mark.unit
def test_asset_with_formula_fails():
    workbook = 'insert_asset_with_formula_and_or_formula_parameters'
    tree = Tree('test_tree_system9', workbook=workbook)

    with pytest.raises(Exception, match="Assets cannot have formulas or formula parameters"):
        tree.insert(pd.DataFrame([{
            'Name': 'Asset with formula',
            'Formula': 'sinusoid()',
            'Type': 'Asset'
        }]), parent='test_tree_system9')

    with pytest.raises(Exception, match="Assets cannot have formulas or formula parameters"):
        tree.insert(pd.DataFrame([{
            'Name': 'Asset with formula',
            'Formula Parameters': '$x=sinusoid()',
            'Type': 'Asset'
        }]), parent='test_tree_system9')

    with pytest.raises(Exception, match="Assets cannot have formulas or formula parameters"):
        tree.insert(pd.DataFrame([{
            'Name': 'Asset with formula',
            'Formula Parameters': '$x=sinusoid()',
            'Formula': '$x',
            'Type': 'Asset'
        }]), parent='test_tree_system9')


@pytest.mark.unit
def test_formula_parameter_pointing_to_asset_fail():
    spy.options.allow_version_mismatch = True
    workbook = 'asset_as_param'
    tree = Tree('test_tree_system11', workbook=workbook)
    tree.insert('Asset1')
    with pytest.raises(RuntimeError, match='Formula parameters must be conditions, scalars, or signals.'):
        tree.insert(pd.DataFrame([{
            'Name': "Fail Due To Asset As Param",
            'Formula Parameters': '$x = Asset1',
            'Formula': '$x'
        }]), parent='test_tree_system11')


@pytest.mark.unit
def test_move():
    tree = Tree('Root')
    tree.insert(['Cooling Tower 1', 'Cooling Tower 2'])
    tree.insert('Area A', parent='Cooling Tower 1')
    tree.insert('Area D', parent='Cooling Tower 2')
    tree.insert(['Temperature', 'Humidity'], parent='Area *')
    tree.move('Area *', 'Cooling Tower 1')

    expected = _build_dataframe_from_path_name_depth_triples([
        ('', 'Root', 1),
        ('Root', 'Cooling Tower 1', 2),
        ('Root >> Cooling Tower 1', 'Area A', 3),
        ('Root >> Cooling Tower 1 >> Area A', 'Humidity', 4),
        ('Root >> Cooling Tower 1 >> Area A', 'Temperature', 4),
        ('Root >> Cooling Tower 1', 'Area D', 3),
        ('Root >> Cooling Tower 1 >> Area D', 'Humidity', 4),
        ('Root >> Cooling Tower 1 >> Area D', 'Temperature', 4),
        ('Root', 'Cooling Tower 2', 2)
    ])

    assert_frame_equal(tree._dataframe, expected)


@pytest.mark.unit
def test_move_errors():
    tree = Tree(pd.DataFrame([{
        'Name': 'Item 1'
    }, {
        'Name': 'Item 2',
        'Path': 'Item 1'
    }, {
        'Name': 'Item 3',
        'Path': 'Item 1 >> Item 2'
    }]))

    with pytest.raises(ValueError, match='Source cannot contain the destination'):
        tree.move('Item 2', 'Item 3')

    with pytest.raises(ValueError, match='Destination must match a single element'):
        tree.move('Item 3', 'Item *')


@pytest.mark.unit
def test_fill_column_values():
    df = pd.DataFrame([{
        'Name': 'Site A_Temp',
        'Unit': '\u00b0F',
        'Facility Type': 'Research Facility'
    }, {
        'Name': 'Site B_Flow',
        'Unit': 'gal/s',
        'Facility Type': 'Factory'
    }])

    queries_to_expected_output = {
        ('{{Name}}',
         '{{Name}}'): pd.Series(['Site A_Temp', 'Site B_Flow']),
        ('{{Name}Site [A-Z]_(.*)}',
         '{{Name}Site ?_(*)}'): pd.Series(['Temp', 'Flow']),
        ('{{Facility Type}} {{Name}Site ([A-Z])_.*}',
         '{{Facility Type}} {{Name}Site (?)_*}'): pd.Series(['Research Facility A', 'Factory B']),
        ('Average of {{Name}(Site [A-Z])_.*} {{Name}Site [A-Z]_(.*)} ({{Unit}})',
         'Average of {{Name}(Site ?)_*} {{Name}Site ?_(*)} ({{Unit}})'): pd.Series(
            ['Average of Site A Temp (\u00b0F)', 'Average of Site B Flow (gal/s)']),
        ('{{Name}(.*_Temp)}',
         '{{Name}(*_Temp)}'): pd.Series(['Site A_Temp', np.nan])
    }

    for (re_query, glob_query), output in queries_to_expected_output.items():
        df['Parent'] = df.apply(_match.fill_column_values, axis=1, query=re_query)
        assert df['Parent'].equals(output)
        df['Parent'] = df.apply(_match.fill_column_values, axis=1, query=glob_query)
        assert df['Parent'].equals(output)


@pytest.mark.unit
def test_insert_with_parent_column():
    expected_df = _build_dataframe_from_path_name_depth_triples([
        ('', 'Root', 1),
        ('Root', 'Area A', 2),
        ('Root >> Area A', 'Optimizer', 3),
        ('Root >> Area A', 'Temperature', 3),
        ('Root', 'Area B', 2),
        ('Root >> Area B', 'Optimizer', 3),
        ('Root >> Area B', 'Temperature', 3),
        ('Root', 'Area C', 2),
        ('Root >> Area C', 'Temperature', 3),
    ])
    tree = Tree('Root')
    tree.insert(['Area A', 'Area B', 'Area C'])
    tree.insert(pd.DataFrame([{
        'Name': 'Optimizer',
        'Type': 'Asset',
        'Parent': r'Area (A|B)'
    }, {
        'Name': 'Temperature',
        'Type': 'Asset',
        'Parent': r'Area \w'
    }]))

    assert_frame_equal(tree._dataframe, expected_df)


@pytest.mark.unit
def test_insert_with_column_values():
    expected_df = _build_dataframe_from_path_name_depth_triples([
        ('', 'Root', 1),
        ('Root', 'Factory B', 2),
        ('Root >> Factory B', 'Flow', 3),
        ('Root', 'Research Facility A', 2),
        ('Root >> Research Facility A', 'Site A_Temp', 3)
    ])

    tree = Tree('Root')
    tree.insert(['Research Facility A', 'Factory B'])
    tree.insert(pd.DataFrame([{
        'Name': 'Site A_Temp',
        'Type': 'Asset',
        'Unit': '\u00b0F',
        'Facility Type': 'Research Facility'
    }, {
        'Name': 'Site B_Flow',
        'Type': 'Asset',
        'Unit': 'gal/s',
        'Facility Type': 'Factory'
    }, {
        'Name': 'This item will not have a matching parent',
        'Type': 'Asset'
    }]), friendly_name='{{Name}Site B_(*)}',
        parent='{{Facility Type}} {{Name}Site (?)_*}')

    assert_frame_equal(tree._dataframe, expected_df)


@pytest.mark.unit
def test_column_values_non_string_data():
    expected_df = _build_dataframe_from_path_name_depth_triples([
        ('', 'Root', 1),
        ('Root', 'Unit 1', 2),
        ('Root', 'Unit True', 2)
    ])

    tree = Tree('Root')
    children = pd.DataFrame()
    children['Name'] = ['beep', 'bop']
    children['Unit Number'] = [1, True]
    tree.insert(children, friendly_name='Unit {{Unit Number}}')

    assert_frame_equal(tree._dataframe, expected_df)


@pytest.mark.unit
def test_insert_parent_dataframe():
    expected_df = _build_dataframe_from_path_name_depth_triples([
        ('', 'Root', 1),
        ('Root', 'Insert Under Me 1st', 2),
        ('Root >> Insert Under Me 1st', '1st Insert', 3),
        ('Root >> Insert Under Me 1st', 'Insert Me Under All', 3),
        ('Root', 'Insert Under Me 2nd', 2),
        ('Root >> Insert Under Me 2nd', '2nd Insert', 3),
        ('Root >> Insert Under Me 2nd', 'Insert Me Under All', 3),
        ('Root', 'Insert Under Me 3rd', 2),
        ('Root >> Insert Under Me 3rd', '3rd Insert', 3),
        ('Root >> Insert Under Me 3rd', 'Insert Me Under All', 3),
        ('Root', 'Insert Under Me 4th', 2),
        ('Root >> Insert Under Me 4th', '4th Insert', 3),
        ('Root >> Insert Under Me 4th', 'Insert Me Under All', 3),
    ])
    sample_id = _common.new_placeholder_guid()
    sample_ref_id = _common.new_placeholder_guid()

    expected_df.loc[expected_df['Name'] == 'Insert Under Me 3rd', 'ID'] = sample_id
    expected_df.loc[expected_df['Name'] == 'Insert Under Me 4th', 'Referenced ID'] = sample_ref_id

    tree = Tree('Root')
    tree.insert([f'Insert Under Me {s}' for s in ('1st', '2nd', '3rd', '4th')])
    # Manually insert ID and Referenced ID -- inserting w/ tree.insert() would require a login
    tree._dataframe.loc[tree._dataframe['Name'] == 'Insert Under Me 3rd', 'ID'] = sample_id
    tree._dataframe.loc[tree._dataframe['Name'] == 'Insert Under Me 4th', 'Referenced ID'] = sample_ref_id

    full_parent_dataframe = pd.DataFrame([{
        # Name case
        'Name': 'Insert Under Me 1st'
    }, {
        # Name + Path case
        'Name': 'Insert Under Me 2nd',
        'Path': 'Root'
    }, {
        # ID case
        'ID': sample_id,
        'Name': 'Doesn\'t matter'
    }, {
        # Referenced ID case
        'ID': sample_ref_id,
        'Name': 'Doesn\'t matter'
    }])
    full_parent_dataframe['Extra Column'] = 'Anything'

    # Perform insertions with `parent` given by a dataframe
    for i, s in enumerate(['1st', '2nd', '3rd', '4th']):
        tree.insert(f'{s} Insert', parent=full_parent_dataframe.iloc[i:i + 1])
    tree.insert('Insert Me Under All', parent=full_parent_dataframe)

    assert_frame_equal(tree._dataframe, expected_df)


@pytest.mark.unit
def test_constructor_with_friendly_name():
    expected = _build_dataframe_from_path_name_depth_triples([
        ('', 'Friendly Root Name', 1),
        ('Friendly Root Name', 'Friendly Item 1 Name', 2),
        ('Friendly Root Name', 'Item 2', 2)
    ])

    data = pd.DataFrame([{
        'Name': 'Root // Friendly Root Name'
    }, {
        'Path': 'Root // Friendly Root Name',
        'Name': 'Item 1 // Friendly Item 1 Name'
    }, {
        'Path': 'Root // Friendly Root Name',
        'Name': 'Item 2'
    }])
    tree = Tree(data, friendly_name='{{Name}*// (*)}')

    assert_frame_equal(tree._dataframe, expected)


@pytest.mark.unit
def test_rename():
    expected = _build_dataframe_from_path_name_depth_triples([
        ('', 'New Root Name', 1),
        ('New Root Name', 'Item 1', 2),
        ('New Root Name >> Item 1', 'Item 2', 3),
        ('New Root Name', 'Item 3', 2)
    ])

    data = pd.DataFrame([{
        'Name': 'Root'
    }, {
        'Path': 'Root',
        'Name': 'Item 1'
    }, {
        'Path': 'Root',
        'Name': 'Item 3'
    }, {
        'Path': 'Root >> Item 1',
        'Name': 'Item 2'
    }])

    tree1 = Tree(data, friendly_name='New Root Name')
    assert_frame_equal(tree1._dataframe, expected)

    tree2 = Tree(data)
    tree2.name = 'New Root Name'
    assert_frame_equal(tree2._dataframe, expected)


@pytest.mark.unit
def test_is_column_value_query():
    assert _match.is_column_value_query('{{Name}}') is True
    assert _match.is_column_value_query('{{Name}(Area ?)_*}') is True
    assert _match.is_column_value_query('{{Facility Type} Factory ([a-zA-Z]{1,20})}') is True
    assert _match.is_column_value_query('{ {Name}(Area ?)_*}') is False
    assert _match.is_column_value_query('{(Area ?)_*{Name}}') is False
    assert _match.is_column_value_query('{Name}') is False
    assert _match.is_column_value_query('Name') is False


@pytest.mark.unit
def test_glob_capture_groups():
    query_asset = '(Area ?)_*'
    query_name = 'Area ?_(*)'

    item = 'Area A_Temperature'

    match = re.match(_match.glob_with_capture_groups_to_regex(query_asset), item)
    assert match is not None
    assert match[1] == 'Area A'

    match = re.match(_match.glob_with_capture_groups_to_regex(query_name), item)
    assert match is not None
    assert match[1] == 'Temperature'

    item = 'Mismatching Item'
    match = re.match(_match.glob_with_capture_groups_to_regex(query_asset), item)
    assert match is None


@pytest.mark.unit
def test_make_paths_from_levels():
    # test that it works when levels but no columns are passed
    df = pd.DataFrame([{
        'Level 1': 'Root',
        'Name': 'Area A',
        'Property': 'Anything',
        'Numerical': 123
    }, {
        'Level 1': 'Root',
        'Level 2': 'Area A',
        'Name': 'Temperature',
        'Property': 'New Value',
        'Numerical': 1
    }, {
        'Level 1': 'Root',
        'Level 2': 'Area A',
        'Name': 'Optimizer',
        'Property': 'Anything',
        'Numerical': 2
    }])
    expected_df = pd.DataFrame([{
        'Path': 'Root',
        'Name': 'Area A',
        'Property': 'Anything',
        'Numerical': 123
    }, {
        'Path': 'Root >> Area A',
        'Name': 'Temperature',
        'Property': 'New Value',
        'Numerical': 1
    }, {
        'Path': 'Root >> Area A',
        'Name': 'Optimizer',
        'Property': 'Anything',
        'Numerical': 2
    }])
    _csv.make_paths_from_levels(df)
    assert_frame_equal(df, expected_df)

    # test that it works when only path is passed
    path = expected_df.copy()
    _csv.make_paths_from_levels(path)
    assert_frame_equal(path, expected_df)

    # test that it works when level and path are provided
    expected_both = pd.DataFrame([{
        'Path': 'Root',
        'Name': 'Area A',
        'Property': 'Anything',
        'Numerical': 123
    }, {
        'Path': 'Root >> Area A',
        'Name': 'Temperature',
        'Property': 'New Value',
        'Numerical': 1
    }, {
        'Path': 'Root >> Area A',
        'Name': 'Optimizer',
        'Property': 'Anything',
        'Numerical': 2
    }])
    both = expected_both.copy()
    _csv.make_paths_from_levels(both)
    assert_frame_equal(both, expected_both)

    # test for warning when neither level nor path is provided
    fail_df = pd.DataFrame([{
        'Name': 'Area A',
        'Property': 'Anything',
        'Numerical': 123
    }, {
        'Name': 'Temperature',
        'Property': 'New Value',
        'Numerical': 1
    }, {
        'Name': 'Optimizer',
        'Property': 'Anything',
        'Numerical': 2
    }])

    with pytest.raises(ValueError):
        _csv.make_paths_from_levels(fail_df)


@pytest.mark.unit
def test_visualize():
    tree = Tree('My Tree')
    tree.insert(['Cooling Tower 1', 'Cooling Tower 2'])
    tree.insert(['Area A', 'Area B', 'Area C'], 'Cooling Tower 1')
    tree.insert('Area D', 'Cooling Tower 2')
    tree.insert(['Temperature', 'Compressor Power', 'Relative Humidity'], 'Area ?')

    expected_full = \
        "My Tree\n" \
        "|-- Cooling Tower 1\n" \
        "|   |-- Area A\n" \
        "|   |   |-- Compressor Power\n" \
        "|   |   |-- Relative Humidity\n" \
        "|   |   |-- Temperature\n" \
        "|   |-- Area B\n" \
        "|   |   |-- Compressor Power\n" \
        "|   |   |-- Relative Humidity\n" \
        "|   |   |-- Temperature\n" \
        "|   |-- Area C\n" \
        "|       |-- Compressor Power\n" \
        "|       |-- Relative Humidity\n" \
        "|       |-- Temperature\n" \
        "|-- Cooling Tower 2\n" \
        "    |-- Area D\n" \
        "        |-- Compressor Power\n" \
        "        |-- Relative Humidity\n" \
        "        |-- Temperature\n"

    expected_subtree = \
        "Cooling Tower 1\n" \
        "|-- Area A\n" \
        "|   |-- Compressor Power\n" \
        "|   |-- Relative Humidity\n" \
        "|   |-- Temperature\n" \
        "|-- Area B\n" \
        "|   |-- Compressor Power\n" \
        "|   |-- Relative Humidity\n" \
        "|   |-- Temperature\n" \
        "|-- Area C\n" \
        "    |-- Compressor Power\n" \
        "    |-- Relative Humidity\n" \
        "    |-- Temperature\n"

    with redirect_stdout(io.StringIO()) as stdout:
        tree.visualize()
    stdout.seek(0)
    output = stdout.read()
    assert output == expected_full

    with redirect_stdout(io.StringIO()) as stdout:
        tree.visualize(subtree='Cooling Tower 1')
    stdout.seek(0)
    output = stdout.read()
    assert output == expected_subtree



import logging
import os

import pytest

from seeq.spy.jobs import _common


@pytest.mark.unit
def test_running_in_datalab():
    os.environ['SEEQ_SDL_CONTAINER_IS_DATALAB'] = ''
    assert _common.running_in_datalab() is False

    os.environ['SEEQ_SDL_CONTAINER_IS_DATALAB'] = 'true'
    assert _common.running_in_datalab() is True


@pytest.mark.unit
def test_running_in_executor():
    os.environ['SEEQ_SDL_CONTAINER_IS_EXECUTOR'] = ''
    assert _common.running_in_executor() is False

    os.environ['SEEQ_SDL_CONTAINER_IS_EXECUTOR'] = 'true'
    assert _common.running_in_executor() is True


@pytest.mark.unit
def test_label_from_executor():
    assert _common.get_label_from_executor() == ''

    os.environ['SEEQ_SDL_LABEL'] = ''
    assert _common.get_label_from_executor() == ''

    os.environ['SEEQ_SDL_LABEL'] = 'explicit'
    assert _common.get_label_from_executor() == 'explicit'


@pytest.mark.unit
def test_log_level_from_executor():
    assert _common.get_log_level_from_executor() == 'INFO'

    os.environ['LOG_LEVEL'] = 'DEBUG'
    assert _common.get_log_level_from_executor() == 'DEBUG'


@pytest.mark.unit
def test_get_executor_logger():
    assert _common.executor_logger is None

    exec_logger = _common.get_executor_logger()
    assert exec_logger.name == "executor_logger"

    assert logging.getLevelName(exec_logger.level) == _common.get_log_level_from_executor()



import mock
import pytest

from seeq import spy
from seeq.sdk import *
from seeq.spy import _login, Session
from seeq.spy import addons
from seeq.spy._errors import *
from seeq.spy.addons import _common as _addons_common

ERROR_52 = f"Seeq R52 versions before R52.1.5 do not support spy.addons"
ERROR_53 = f"Seeq R53 versions before R53.0.2 do not support spy.addons"


def enable_addon_tools(session: Session, value):
    system_api = SystemApi(session.client)
    config_option_input = ConfigurationOptionInputV1(path='Features/AddOnTools/Enabled', value=value)
    system_api.set_configuration_options(body=ConfigurationInputV1([config_option_input]))


def test_addon_support(session: Session, addon_function, input_args):
    # only testing for the current major version. Otherwise, _login.validate will complain of different
    # SPy-SeeqServer versions
    major_version, *_ = _login.get_server_version_tuple(spy.session)
    error = ERROR_52 if major_version == 52 else ERROR_53 if major_version == 53 else None
    if major_version <= 53:
        with mock.patch.object(_login, 'get_server_version_tuple', return_value=(major_version, 0, 0)):
            with pytest.raises(SPyException, match=error):
                getattr(addons, addon_function)(input_args)


@pytest.mark.unit
def test_verify_addons_support():
    with mock.patch.object(_login, 'get_server_version_tuple', return_value=(52, 0, 0)):
        with pytest.raises(SPyException, match=ERROR_52):
            _addons_common.verify_addons_support(spy.session)

    with mock.patch.object(_login, 'get_server_version_tuple', return_value=(52, 1, 4)):
        with pytest.raises(SPyException, match=ERROR_52):
            _addons_common.verify_addons_support(spy.session)

    with mock.patch.object(_login, 'get_server_version_tuple', return_value=(52, 1, 5)):
        _addons_common.verify_addons_support(spy.session)

    with mock.patch.object(_login, 'get_server_version_tuple', return_value=(52, 2, 0)):
        _addons_common.verify_addons_support(spy.session)

    with mock.patch.object(_login, 'get_server_version_tuple', return_value=(53, 0, 0)):
        with pytest.raises(SPyException, match=ERROR_53):
            _addons_common.verify_addons_support(spy.session)

    with mock.patch.object(_login, 'get_server_version_tuple', return_value=(53, 0, 1)):
        with pytest.raises(SPyException, match=ERROR_53):
            _addons_common.verify_addons_support(spy.session)

    with mock.patch.object(_login, 'get_server_version_tuple', return_value=(53, 0, 2)):
        _addons_common.verify_addons_support(spy.session)

    with mock.patch.object(_login, 'get_server_version_tuple', return_value=(53, 1, 0)):
        _addons_common.verify_addons_support(spy.session)

    with mock.patch.object(_login, 'get_server_version_tuple', return_value=(54, 1, 0)):
        _addons_common.verify_addons_support(spy.session)



import os

from dataclasses import dataclass
from typing import Optional, List

from seeq import spy
from seeq.sdk import *
from seeq.sdk.configuration import ClientConfiguration
from seeq.spy import _url
from seeq.spy._config import Options

SEEQ_SDK_RETRY_TIMEOUT_IN_SECONDS_ENV_VAR_NAME = 'SEEQ_SDK_RETRY_TIMEOUT_IN_SECONDS'


@dataclass(repr=False)
class Session:
    """
    Used to segregate Seeq Server logins and allows for multi-server /
    multi-user concurrent logins. This object encapsulates all server-
    specific state, SPy options and API client configuration.

    Examples
    --------
    Log in to two different servers at the same time:

    >>> session1 = Session()
    >>> session2 = Session()
    >>> spy.login(url='https://server1.seeq.site', username='mark', password='markpassword', session=session1)
    >>> spy.login(url='https://server2.seeq.site', username='alex', password='alexpassword', session=session2)
    """
    options: Options = None
    client_configuration: ClientConfiguration = None
    _client: Optional[ApiClient] = None
    _user: Optional[UserOutputV1] = None
    _public_url: Optional[str] = None
    _private_url: Optional[str] = None
    _server_version: Optional[str] = None
    supported_units: Optional[set] = None
    corporate_folder: Optional[FolderOutputV1] = None
    auth_providers: Optional[List[DatasourceOutputV1]] = None
    https_verify_ssl: bool = True
    https_key_file: Optional[str] = None
    https_cert_file: Optional[str] = None

    def __init__(self, options: Options = None, client_configuration: ClientConfiguration = None):
        self.client_configuration = client_configuration if client_configuration is not None else ClientConfiguration()
        self.options = options if options is not None else Options(self.client_configuration)

        # We have this mechanism so that test_run_notebooks() is able to increase the timeout for the child kernels
        if Session.get_global_sdk_retry_timeout_in_seconds() is not None:
            self.options.retry_timeout_in_seconds = Session.get_global_sdk_retry_timeout_in_seconds()

    def __repr__(self):
        if not self.client:
            return 'Not logged in'
        url_part = self.public_url
        if self.private_url != self.public_url:
            url_part += f' ({self.private_url})'
        return f'{url_part} as {self.user.name} ({self.user.username})'

    def __getstate__(self):
        # We can only pickle certain members. This has to mirror __setstate__().
        return self.options

    def __setstate__(self, state):
        self.options = state

    @staticmethod
    def validate(session):
        return spy.session if session is None else session

    @staticmethod
    def set_global_sdk_retry_timeout_in_seconds(timeout: Optional[int]):
        """
        This is used to set the SDK's retry timeout (see
        "retry_timeout_in_seconds" in api_client.py) for all
        child Python kernels, such as those spawned by executing
        notebooks via nbformat is in test_run_notebook().
        :param timeout: Timeout (in seconds)
        """
        if timeout is None and SEEQ_SDK_RETRY_TIMEOUT_IN_SECONDS_ENV_VAR_NAME in os.environ:
            del os.environ[SEEQ_SDK_RETRY_TIMEOUT_IN_SECONDS_ENV_VAR_NAME]
        else:
            os.environ[SEEQ_SDK_RETRY_TIMEOUT_IN_SECONDS_ENV_VAR_NAME] = str(timeout)

    @staticmethod
    def get_global_sdk_retry_timeout_in_seconds() -> Optional[int]:
        """
        See set_global_sdk_retry_timeout_in_seconds()
        :return: Timeout (in seconds)
        """
        if SEEQ_SDK_RETRY_TIMEOUT_IN_SECONDS_ENV_VAR_NAME in os.environ:
            return int(os.environ[SEEQ_SDK_RETRY_TIMEOUT_IN_SECONDS_ENV_VAR_NAME])
        else:
            return None

    def clear(self):
        """
        Re-initializes the object to a "logged out" state. Note that this
        function does NOT reset API client configuration or SPy options.
        """
        self.client = None
        self.user = None
        self.public_url = None
        self.private_url = None
        self.server_version = None
        self.supported_units = None
        self.corporate_folder = None
        self.auth_providers = None
        self.https_verify_ssl = True
        self.https_key_file = None
        self.https_cert_file = None

    # Prior to the advent of Session objects, the spy.client, spy.user and spy.server_version module-level variables
    # were exposed to end-users as a convenience. The setters below copy those (now) Session variables to those
    # legacy module-level locations for backward compatibility purposes. (Only if this Session object is the default
    # Session.)
    @property
    def client(self):
        return self._client

    @client.setter
    def client(self, value):
        self._client = value
        if self is spy.session:
            spy.client = self._client

    @property
    def user(self):
        return self._user

    @user.setter
    def user(self, value):
        self._user = value
        if self is spy.session:
            spy.user = self._user

    @property
    def server_version(self):
        return self._server_version

    @server_version.setter
    def server_version(self, value):
        self._server_version = value
        if self is spy.session:
            spy.server_version = self._server_version

    @property
    def public_url(self):
        return self._public_url

    @public_url.setter
    def public_url(self, value):
        self._public_url = _url.cleanse_url(value)

    @property
    def private_url(self):
        return self._private_url

    @private_url.setter
    def private_url(self, value):
        self._private_url = _url.cleanse_url(value)

    def get_api_url(self):
        """
        Returns the URL to use for API calls, which ends up being the
        private URL (if specified) or the public URL.
        """
        return f'{self.private_url}/api'



from packaging import version

from seeq.spy import _login
from seeq.spy._errors import *
from seeq.spy._session import Session


def verify_addons_support(session: Session):
    # We will have to keep this code till R53 is phased out, since there is a possibility of having versions R52 or
    # R53 with the external_tools endpoint.
    seeq_server_major, seeq_server_minor, seeq_server_patch = _login.get_server_version_tuple(session)
    seeq_server_version = f"{seeq_server_major}.{seeq_server_minor}.{seeq_server_patch}"
    if seeq_server_major == 52:
        if version.parse(seeq_server_version) < version.parse("52.1.5"):
            raise SPyValueError(f"Seeq R52 versions before R52.1.5 do not support spy.addons")
    elif seeq_server_major == 53:
        if version.parse(seeq_server_version) < version.parse("53.0.2"):
            raise SPyValueError(f"Seeq R53 versions before R53.0.2 do not support spy.addons")



"""
  Implements utility methods used to determine information and run commands across different platforms.
"""
import os
import pathlib
import re
import threading

from typing import Optional, Tuple, Union, Iterable


# See http://code.activestate.com/recipes/52308-the-simple-but-handy-collector-of-a-bunch-of-named
class Bunch:
    def __init__(self, **kwds):
        self.__dict__.update(kwds)


UNITS_BASE_TWO = ["B", "KiB", "MiB", "GiB", "TiB", "PiB", "EiB"]
UNITS_BASE_TEN = ["B", "KB", "MB", "GB", "TB", "PB", "EB"]
BYTES_PER_KIBIBYTE = 1024
BYTES_PER_MEBIBYTE = BYTES_PER_KIBIBYTE * BYTES_PER_KIBIBYTE
BYTES_PER_MEGABYTE = 1e6
BYTES_PER_GIGABYTE = 1e9


def get_system():
    import platform
    return platform.system()


def get_release():
    import platform
    return platform.release()


def get_architecture():
    import platform
    return platform.architecture()


def get_platform():
    import platform
    plat = platform.platform()
    if plat.lower().find('windows') != -1:
        return 'windows'
    if plat.lower().find('linux') != -1:
        return 'linux'
    if plat.lower().find('darwin') != -1:
        return 'osx'


def get_linux_distribution():
    import platform
    # Examples of platform.platform() return values:
    # Ubuntu 18.10 on Azure:  Linux-5.0.0-1027-azure-x86_64-with-debian-buster-sid

    if not is_linux():
        return None

    platform_map = {
        'debian': 'debian',
        'ubuntu': 'debian',
        'fedora': 'fedora',
        'redhat': 'fedora',
        'centos': 'fedora'
    }

    plat = platform.platform().lower()
    for fragment, distribution in platform_map.items():
        if fragment in plat:
            return distribution

    return None


def get_system_bit():
    if is_64_bit():
        return '64bit'
    else:
        return '32bit'


def get_build_architecture():
    return os.environ['SQ_ARCHITECTURE']


def is_build_x86():
    return get_build_architecture() == 'x86'


def file_empty(path):
    apath = os.path.join(get_project_root_dir(), path)
    return not os.path.isfile(apath) or os.stat(apath).st_size == 0


def print_file(path, indent=''):
    num_lines = 0
    with open(os.path.join(get_project_root_dir(), path)) as f:
        for line in f.readlines():
            print((indent + line))
            num_lines += 1
    return num_lines


def is_windows():
    return get_platform() == 'windows'


def is_linux():
    return get_platform() == 'linux'


def is_osx():
    return get_platform() == 'osx'


g_repo_root_dir = None


def get_repo_root_dir(windows_long_path=True):
    """
    Returns the path to the root of the repository. i.e., C:\crab\
    """
    global g_repo_root_dir
    if not g_repo_root_dir:
        # Cache it since this is an expensive call
        dir1 = os.path.abspath(os.path.dirname(__file__))
        while not os.path.exists(os.path.join(dir1, 'repo-root-dir.txt')):
            dir2 = os.path.abspath(os.path.join(dir1, os.pardir))
            if dir1 == dir2:
                # We hit the root and didn't find it
                return

            dir1 = dir2

        g_repo_root_dir = dir1

    return cleanse_path(g_repo_root_dir, windows_long_path)


def get_git_dir(windows_long_path=True):
    """
    In the most common case, returns the path to the to the repository's
    '.git' folder. i.e., C:\crab\.git
    """
    return cleanse_path(os.path.abspath(git_rev_parse('--git-dir')), windows_long_path)


def get_git_common_dir(windows_long_path=True):
    """
    Returns the path to the '.git' folder. In git worktrees the
    '--git-dir' flag returns a path to a text file. In a worktree
    the '--git-common-dir' flag returns the path to the true
    '.git' directory in the main git worktree
    """
    # noinspection PyBroadException
    try:
        # common dir is incorrect when ran from a sub folder
        # See https://goo.gl/MTHFQf
        wd = get_repo_root_dir(windows_long_path)
        return cleanse_path(os.path.join(wd, git_rev_parse('--git-common-dir', cwd=wd)), windows_long_path)
    except BaseException:
        # We are likely on an older version of git, use the --git-dir flag (which has existed longer)
        return get_git_dir(windows_long_path)


def get_project_name():
    project_root_dir = get_project_root_dir()

    # Get rid of everything before the last forward slash
    return project_root_dir.split(os.sep)[-1]


g_project_root_dir = ''


def get_project_root_dir(windows_long_path=True):
    if g_project_root_dir == '':
        raise Exception('g_project_root_dir not set')

    project_root_dir = cleanse_path(g_project_root_dir, windows_long_path)

    return project_root_dir


def set_project_root_dir(project_root_dir):
    global g_project_root_dir
    g_project_root_dir = project_root_dir


def get_common_dir(windows_long_path=True):
    common_dir = os.path.join(g_project_root_dir, '..', 'common')
    common_dir = os.path.normpath(common_dir)

    if not os.path.exists(common_dir):
        common_dir = os.path.join(g_project_root_dir, 'common')

    if not os.path.exists(common_dir):
        raise Exception('common directory could not be found')

    return cleanse_path(common_dir, windows_long_path)


def get_home_dir(user=None):
    if is_windows():
        home = os.environ['USERPROFILE']
    else:
        user_to_expand = user if user else ''
        home = os.path.expanduser(f'~{user_to_expand}')

    return home


def get_sq_dir():
    return os.environ['SQ_FOLDER']


def is_64_bit():
    if is_windows():
        # On Windows, even if Python is running as a 32-bit process, we are actually
        # more interested in whether we are on a 64-bit flavor of Windows.
        # In a 32-bit process on 64-bit Windows, PROCESSOR_ARCHITEW6432 will be AMD64.
        # In a 64-bit process on 64-bit Windows, PROCESSOR_ARCHITECTURE will be AMD64.
        return os.environ.get('PROCESSOR_ARCHITEW6432') == 'AMD64' or os.environ.get(
            'PROCESSOR_ARCHITECTURE') == 'AMD64'
    else:
        import struct
        # On UNIX operating systems, Python will be either 32- or 64-bit according to the
        # operating system. So we can just look at the size of a pointer.
        return True if struct.calcsize("P") * 8 == 64 else False


def is_os_headless():
    """
      Returns True if a display is not detected in the build environment
    """
    return get_platform() == 'linux' and 'DISPLAY' not in os.environ


def render_command_line_posix(cmd_and_args: Iterable[Union[str, bytes]]) -> str:
    """
    Render a command line from its program + argument list into a form that can be consumed by a POSIX-like
    shell (sh, bash, zsh, etc) in which each argument is preserved unambiguously.

    >>> print(render_command_line_posix(['find', 'Cool Beans/', '-iname', '*.txt']))
    find 'Cool Beans/' -iname '*.txt'
    """
    import shlex
    return ' '.join(shlex.quote(os.fsdecode(word)) for word in cmd_and_args)


def render_command_line_msvcrt(cmd_and_args: Iterable[Union[str, bytes]]) -> str:
    """
    Render a command line from its program + argument list into a form that can be interpreted by the MSVCRT on Windows,
    for which each argument is preserved unambiguously.

    Note that this does NOT suppress cmd.exe-specific syntax that might be applied prior to execution, if run that way,
    e.g. environment variable expansions: it just surrounds tokens by quotes as necessary to group them.

    >>> print(render_command_line_msvcrt(['dir', '/q', 'Some Folder']))
    dir /q "Some Folder"
    """
    from subprocess import list2cmdline
    return list2cmdline([os.fsdecode(word) for word in cmd_and_args])


def render_command_line(cmd_and_args: Iterable[Union[str, bytes]]) -> str:
    """
    Render a command line from its program + argument list for the current platform.
    The quoting is performed as a command for `cmd.exe` on Windows, or a POSIX-ish shell otherwise.

    See also `render_command_line_msvcrt` and `render_command_line_posix`.
    """
    if is_windows():
        return render_command_line_msvcrt(cmd_and_args)
    else:
        return render_command_line_posix(cmd_and_args)


def spawn(
        command_line,
        *,
        raise_on_error=True,
        capture_output=False,
        wait_for_end=True,
        suppress_input=False,
        suppress_output=False,
        suppress_error=False,
        new_process_group=False,
        cwd=None,
        virtual_display_if_necessary=False,
        env=None,
        echo=False,
        shell: Optional[bool] = None,
        stdout=None,
        stderr=None):
    """
      Runs the 'command_line' input in the platform's command line.
    """
    import subprocess
    from subprocess import PIPE
    import sys
    additional_kargs = dict()

    if not cwd:
        # noinspection PyBroadException
        try:
            cwd = get_project_root_dir(windows_long_path=False)
        except BaseException:
            cwd = os.getcwd()

    if env is None:
        env = os.environ

    stdin = None

    if not stderr:
        # By default, we pipe stderr into stdout because Bamboo buffers stdout and not stderr, so you
        # can get out-of-order logs that are really confusing.
        # See https://answers.atlassian.com/questions/31595/bamboo-build-log-not-in-sync
        stderr = subprocess.STDOUT

    output = ''
    err = ''

    if capture_output and suppress_output:
        raise Exception('capture_output and suppress_output are mutually exclusive.')

    if capture_output:
        stdout = PIPE
        stderr = PIPE

    if suppress_input:
        stdin = open(os.devnull, 'r')

    if suppress_output:
        stdout = open(os.devnull, 'w')

    if suppress_error:
        stderr = open(os.devnull, 'w')

    # One reason you might want to do this is to prevent a CTRL^C in the parent process from killing the child
    # process so that the child process can be shutdown gracefully
    if new_process_group:
        # Based on https://stackoverflow.com/questions/5045771
        if is_windows():
            # https://docs.microsoft.com/en-us/windows/win32/procthread/process-creation-flags
            additional_kargs['creationflags'] = 0x00000200
        else:
            def preexec_fn():
                os.setpgrp()

            additional_kargs['preexec_fn'] = preexec_fn

    # Detect if we are running headless and if so run using xvfb
    if virtual_display_if_necessary and is_os_headless():
        prefix = ['xvfb-run', '--auto-servernum', '--server-args=-screen 0 1024x768x24']
        if isinstance(command_line, (str, bytes)):
            prefix = f'{render_command_line(prefix)} '
            prefix = os.fsencode(prefix) if isinstance(command_line, bytes) else prefix
        command_line = prefix + command_line

    if echo:
        print(command_line)

    if shell is None:
        # Assume that if the command is a string (or bytes), then it's a pre-formatted command line to be interpreted by
        # cmd.exe or sh. Otherwise run as a proper exec command ([program, argv1, argv2, …]).
        # Although you *can* use the shell with an argument list, that is a slightly more advanced mode of operation,
        # and behaves wildly differently for windows vs. posix, and so you should pass `shell=True` explicitly
        # if you want that.
        shell = isinstance(command_line, (str, bytes))

    proc = subprocess.Popen(command_line, env=env, cwd=cwd, shell=shell, stdin=stdin, stdout=stdout, stderr=stderr,
                            **additional_kargs)

    if capture_output:
        (output, err) = proc.communicate()

    if wait_for_end:
        ret_code = proc.wait()

        if ret_code != 0 and raise_on_error:
            raise RuntimeError('Command failed with exit code %d: %s' %
                               (ret_code, command_line))

        if not isinstance(output, str):
            output = str(output, encoding=sys.stdout.encoding)

        if not isinstance(err, str):
            err = str(err, encoding=sys.stderr.encoding)

        return ret_code, output, err

    else:
        return proc


def git_rev_parse(flag, cwd=None):
    """
    Calls the 'git rev-parse' command with the supplied flag. This is useful for
    getting a path from git. For example, 'git rev-parse --git-dir' will return
    the directory of the '.git' directory.
    """
    (ret_code, output, err) = spawn('git rev-parse --no-flags ' + flag, capture_output=True, cwd=cwd)
    result = output.strip()
    if result == '':
        raise Exception('git rev-parse failed to return a path')
    return result


def convert_windows_path_to_unix(path):
    return re.sub(r'\\', '/', path)


def cleanse_path(path: Union[str, pathlib.Path], windows_long_path: bool = True):
    """
    Takes a potentially 'mixed-platform' path (i.e., partially Windows, partially Unix)
    and turns it into a path that is guaranteed to work on the host platform. Also,
    on Windows it will optionally turn it into a 'long path' that can support >65k
    characters. (This is necessary especially for doing any operations on the
    node_modules folder.)
    """

    # Make sure it is a string (could have been a pathlib.Path)
    path = str(path)

    # Get rid of extra/unnecessary /../ or /./ sections of the path. Note that normpath does not work properly with
    # long filenames, so we have to convert to short paths and then call this before cleaning.
    path = os.path.normpath(re.sub(r'^([/\\][/\\]\?[/\\])?(\w:)', r'\2', path))

    if is_windows():
        # First replace all forward slashes with backslashes like Windows wants.
        path = re.sub(r'/', r'\\', path)
        if windows_long_path:
            # Windows 'long paths' are of the form \\?\D:\Foo\Bar (instead of the
            # regular D:\Foo\Bar). So we have to do a RegEx replacement that involves
            # A LOT of backslashes.
            path = re.sub(r'^(\w):\\', r'\\\\?\\\1:\\', path)
        else:
            # Get rid of the 'long path' form if it's in there. There are a some
            # APIs/utilities that can't handle them.
            path = re.sub(r'^(\\\\\?\\)?(\w:)', r'\2', path)
    else:
        path = convert_windows_path_to_unix(path)

    return path


def cleanse_filename(filename, replacement_char='_'):
    return re.sub(r'[:"%\\/<>^|?*&\[\]]', replacement_char, filename)


def concat_path(path1, path2):
    return cleanse_path(path1 + os.sep + path2)


def script_path_to_unix(path, retain_colon=False):
    """
    This function is used to convert a path into something that is suitable for
    a Unix script, including a Bash script being run on Windows in Git Bash.

    On Windows, a path like r"C:\Foo\Bar" or r"\\?\C:\Foo\Bar" will get turned into
    "/c/foo/bar".
    """
    path = path.replace('\\', '/')
    path = re.sub(r'^(//\?/)?(\w):', '/\\2%s' % (':' if retain_colon else ''), path)
    return path


def script_path_to_windows(path):
    """
    This function is used to convert a path into something that is suitable for
    a Windows batch script.

    On Windows, a path like "C:/Foo/Bar"  will get turned into
    'C:\foo\bar'. Note that the long path prefix ('\\?\') will also get stripped
    off because it's not supported in Windows batch scripts. (Go figure.)
    """
    path = path.replace('/', '\\')
    path = re.sub(r'^(\\\\\?\\)?(\w:)', '\\2', path)
    return path


def path_to_regex(path):
    """
    This function returns a regular expression that will match the given path
    properly against a string that may contain a variety of representations of
    the path on Windows.
    """

    remaining_path = path
    regex = r''

    # On Windows, paths may be prefixed in the following ways:
    # Regular:      C:\
    # Long Paths:   \\?\C:\
    # Git Bash:     /C/

    # The following RegEx extracts the drive letter no matter which of the
    # ways happens to be used
    match = re.search(r'^((\\\\\?\\)?|/)(\w)(:[\\/]|/)', path)
    if match:
        drive_letter = match.group(3)
        remaining_path = path[match.end():]

        # We use an almost identical RegEx for our pattern except we hard-code
        # the drive letter
        regex = r'((\\\\\?\\)?|/)%s(\:[\\/]|/)' % drive_letter

    # Windows is loosey-goosey in that it allows a mixture of backslashes and
    # forward slashes throughout the path, so we replace all of them with
    # a character matching set that includes both
    regex += re.sub(r'[\\/]', r'[\\\\/]', remaining_path)

    # The path could be wrapped in quotes
    regex = '[\'"]?' + regex + '[\'"]?'

    return regex


def find_files_recursively(starting_dir, wildcard):
    import fnmatch
    """
    Returns a list of files that match the wildcard passed in (honoring asterisks)
    by recursively searching the starting_dir and below.
    """
    matches = []
    for root, dirnames, filenames in os.walk(starting_dir):
        for filename in fnmatch.filter(filenames, wildcard):
            matches.append(cleanse_path(os.path.join(root, filename), False))
    return matches


def find_folders_recursively(starting_dir, wildcard):
    import fnmatch
    """
    Returns a list of folders that match the wildcard passed in (honoring asterisks)
    by recursively searching the starting_dir and below.
    """
    matches = []
    for root, dirnames, filenames in os.walk(starting_dir):
        for folder in fnmatch.filter(dirnames, wildcard):
            matches.append(cleanse_path(os.path.join(root, folder), False))
    return matches


def replace_in_string(_str, replacements):
    """
    Replaces all instances of a regex in a string with the specified replacement.
    'replacements' is a list of tuples in the form (pattern, repl, [flags]) where 'pattern'
    is a regex pattern and 'repl' is either a function that, when called with the
    match object returns the replacement text, or a straight regex string that
    represents the replacement. 'flags' is an optional modifier for the regex
    such as re.MULTILINE
    """

    from inspect import isfunction
    while True:
        original = _str
        for replacement in replacements:
            pattern = replacement[0]
            repl = replacement[1]
            flags = replacement[2] if len(replacement) is 3 else 0
            compiled = re.compile(pattern, flags)
            match = compiled.search(_str)
            if match:
                if isfunction(repl):
                    replacement_text = repl(match)
                    _str = _str[:match.start()] + replacement_text + _str[match.end():]
                else:
                    _str = re.sub(pattern, repl, _str, 0, flags)

        if original == _str:
            break

    return _str


def replace_in_file(file_path, replacements):
    """
    Replaces all instances of a regex in a text file with the specified replacement.
    'replacements' is a list of tuples in the form (pattern, repl, [flags]) where 'pattern'
    is a regex pattern and 'repl' is either a function that, when called with the
    match object returns the replacement text, or a straight regex string that
    represents the replacement. 'flags' is an optional modifier for the regex
    such as re.MULTILINE
    """
    file = open(file_path, 'r', encoding='utf-8')
    file_content = file.read()
    file.close()

    file_content = replace_in_string(file_content, replacements)

    file = open(file_path, 'w', encoding='utf-8')
    file.write(file_content)
    file.close()


def replace_in_binary_file(file_path, replacements):
    """
    Replaces all instances of a pattern in a binary file with the specified replacement.
    'replacements' is a list of tuples in the form (pattern, repl) where 'pattern'
    is a pattern and 'repl' is its replacement. Regex is not supported.
    """
    file = open(file_path, 'rb')
    file_content = file.read()
    file.close()

    for replacement in replacements:
        (pattern, repl) = replacement
        file_content = file_content.replace(pattern, repl)

    file = open(file_path, 'wb')
    file.write(file_content)
    file.close()


def get_username():
    """
    Returns the current user's name in a cross platform way.
    """
    if is_windows():
        import win32api
        return win32api.GetUserName()
    else:
        import pwd
        return pwd.getpwuid(os.getuid()).pw_name


def keypress_event():
    """
    Returns a threading.Event object that is signaled if the user presses ENTER.
    """
    event = threading.Event()
    threading.Thread(target=_wait_for_keypress_thread, daemon=True, args=(event,)).start()
    return event


def _wait_for_keypress_thread(event):
    try:
        input()
    except EOFError:
        # If this occurs, then there is no input stream to read from. We exit this thread
        # and the original thread will wait forever (or until its timeout) for the event
        # to be set.
        #
        # This is the case for 'sq run' during 'sq build' for AppServer: The integration
        # test target manager kills 'sq run' using 'sq run --kill', not a keypress.
        return

    event.set()


def is_script_running_on_the_continuous_integration_server():
    return 'IS_CI' in os.environ


def copytree(src, dst, symlinks=False, update=False, mirror=False, permissions=False, progress=False, exclude=None):
    if is_windows():
        robocopy(src, dst, update=update, mirror=mirror, permissions=permissions, progress=progress, exclude=exclude)
    else:
        rsync(src, dst, symlinks=symlinks, update=update, mirror=mirror, permissions=permissions, progress=progress,
              exclude=exclude)


def copyfile(src, dst):
    import stat
    import shutil
    if not os.path.exists(os.path.dirname(dst)):
        os.makedirs(os.path.dirname(dst))

    try:
        shutil.copy2(src, dst)
    except IOError as e:
        if e.errno == 13:  # Permission denied
            # Remove the read-only flag and try again
            os.chmod(dst, stat.S_IWRITE)
            shutil.copy2(src, dst)
        else:
            raise e


def on_rm_error(func, path, exc_info):
    # From
    # http://stackoverflow.com/questions/4829043/how-to-remove-read-only-attrib-directory-with-python-in-windows

    # path contains the path of the file that couldn't be removed
    # let's just assume that it's read-only and unlink it.
    import stat
    os.chmod(path, stat.S_IWRITE)
    os.unlink(path)


def copy_windows_acls(src, dst):
    import tempfile
    import codecs
    # From https://confidentialfiles.wordpress.com/2014/03/13/copying-ntfs-permissions-between-folders/

    # Create a temporary file for the ACL descriptors.
    temp_handle, temp_path = tempfile.mkstemp()
    os.close(temp_handle)

    # Write the ACL descriptors to the temp file.
    spawn('icacls "%s" /save "%s"' % (src, temp_path))

    # Open up the ACL descriptor file. It is encoded as UCS-2 Little Endian.
    with codecs.open(temp_path, 'r', 'utf_16_le') as f:
        lines = f.readlines()

    # Replace the first line, which is the name of the source directory, with the destination directory.
    lines[0] = str(os.path.basename(dst))

    # Write out the new ACL descriptor file.
    with codecs.open(temp_path, 'w', 'utf_16_le') as f:
        f.writelines(os.linesep.join(lines) + os.linesep)

    # Grab the directory name of the destination because (and this is so cool -- not!!) icacls restores ACLs by
    # using the parent directory as the command line input and then the target directory on the first line of the file.
    dst_dir = os.path.dirname(dst)

    # What a pain in the butt! icacls chokes if you pass in "C:\" (with surrounding quotes) for root folders.
    # So only wrap in quotes if there is actually a space somewhere in the directory name.
    if ' ' in dst_dir:
        dst_dir = '"%s"' % dst_dir

    # Finally, "restore" the ACLs to the new location
    spawn('icacls %s /restore "%s"' % (dst_dir, temp_path))

    # Clean up the temp file.
    os.remove(temp_path)


def robocopy(src, dst, mirror=False, update=False, permissions=False, progress=False, exclude=None):
    # Arguments cannot use long path names
    src = cleanse_path(src, windows_long_path=False)
    dst = cleanse_path(dst, windows_long_path=False)
    mode = '/MIR' if mirror else '/E'
    upd = '/XO' if update else ''
    logging = '/NP /NS /NC /NFL /NDL /NJH /NJS'
    if progress:
        logging = '/NP /NS /NC /NJH /NJS'
    exclusions = ''
    if exclude is not None:
        files, dirs = map(lambda ps: ' '.join(ps), _partition(
            lambda path: os.path.isdir(path),
            # get all the existing files that match the exclude path on the src or dst
            (os.path.join(base, x) for x in exclude for base in [src, dst] if os.path.exists(os.path.join(base, x)))
        ))
        if len(files):
            exclusions += ' /XF ' + files
        if len(dirs):
            exclusions += ' /XD ' + dirs
    cmdline = 'robocopy "%s" "%s" %s %s /R:0 /MT %s %s' % (src, dst, mode, upd, logging, exclusions)
    (ret_code, output, err) = spawn(cmdline, capture_output=not progress, raise_on_error=False)
    if ret_code > 3:
        raise Exception(
            'Robocopy failed to copy "%s" to "%s" with error code %d:\n%s\n%s' % (src, dst, ret_code, output, err))

    # Why not use the /SEC switch on robocopy? Because it doesn't seem to work in Windows Server 2008 R2. CRAB-12017
    if permissions:
        copy_windows_acls(src, dst)


def _partition(pred, iterable):
    # Taken from itertools-more (MIT Licensed). It is problematic to add dependencies to seeq.base because they wont
    # be installed early enough and it would be an additional required dependency for spy
    # https://more-itertools.readthedocs.io/en/stable/_modules/more_itertools/recipes.html#partition

    import itertools
    t1, t2 = itertools.tee(((pred(x), x) for x in iterable))
    return (
        (x for (cond, x) in t1 if not cond),
        (x for (cond, x) in t2 if cond),
    )


def rsync(src, dst, symlinks=False, mirror=False, update=False, permissions=False, progress=False, exclude=None):
    if not os.path.exists(dst):
        os.makedirs(dst)

    args = '-rtE'
    if symlinks:
        args = args + 'l'

    if update:
        args = args + 'u'

    if permissions:
        args = args + 'p'

    if not progress:
        args = args + 'q'

    if mirror:
        args = args + ' --del'

    if exclude is not None:
        args = args + ' ' + ' '.join([('--exclude %s' % x) for x in exclude])

    # Add a trailing slash so that rsync won't create an additional level below the destination
    if not src.endswith('/'):
        src = src + '/'

    cmdline = 'rsync %s "%s" "%s"' % (args, src, dst)
    spawn(cmdline)


def do_with_retry(func, timeout_sec=5, retry_sleep_sec=0.1):
    import time
    timer = time.time()
    while True:
        # noinspection PyBroadException
        try:
            func()
            break
        except BaseException:
            if time.time() - timer > timeout_sec:
                raise

            time.sleep(retry_sleep_sec)


def removetree(path, keep_top_folder=False, exclude_subdirectories=None):
    import shutil
    import tempfile
    if exclude_subdirectories is None:
        exclude_subdirectories = list()

    with tempfile.TemporaryDirectory() as empty_folder:
        # If we're on Windows, and there is a folder with contents where part of the path >260 characters, we must
        # use Robocopy (via copytree) to remove the contents of the folder due to extremely long path names.
        # Ironically, however, the top-level folder can't be long-named.
        copytree(str(empty_folder), path, mirror=True, exclude=exclude_subdirectories)

    if not keep_top_folder and len(exclude_subdirectories) == 0:
        # the robocopy process can hold on to the directory we're trying to delete for a short time.
        do_with_retry(lambda: shutil.rmtree(path, onerror=on_rm_error))


def remove_deep_folder(path):
    removetree(path)


def remove_files_from_list(file_list: list):
    for file_path in file_list:
        try:
            os.remove(file_path)
        except OSError:
            print("Error while deleting file : ", file_path)


def get_git_branch():
    """
    Return the name of the current git branch.

    Taken from:
    http://stackoverflow.com/questions/1593051/how-to-programmatically-determine-the-current-checked-out-git-branch
    :return: Name of the current git branch or None if currently in a 'detached HEAD' state
    """
    ret_code, output, err = spawn('git symbolic-ref HEAD', capture_output=True, raise_on_error=False)
    if ret_code != 0:
        if 'ref HEAD is not a symbolic ref' in err:
            return None  # Detached head, no branch
        message = err.replace('fatal: ', '').strip()
        raise RuntimeError(f'Unable to get git branch: {message}')
    return output.replace('refs/heads/', '').strip()


def get_aws_metadata(metadata):
    """
    Return information about the AWS instance that this code is running on.

    Taken from:
    http://docs.aws.amazon.com/AWSEC2/latest/UserGuide/ec2-instance-metadata.html
    :param metadata: The key for the metadata you want to retrieve
    :return: The desired metadata
    """
    import urllib.request
    import urllib.parse
    url = 'http://169.254.169.254/latest/meta-data/%s' % metadata
    with urllib.request.urlopen(url, timeout=1) as f:
        return f.read().decode('utf-8')


def get_current_user_name_and_email() -> Tuple[Optional[str], Optional[str]]:
    user_name = None
    user_email = None

    # noinspection PyBroadException
    try:
        (ret_code, output, err) = spawn('git config user.name', capture_output=True)
        user_name = output.strip() if len(output.strip()) > 0 else None
        (ret_code, output, err) = spawn('git config user.email', capture_output=True)
        user_email = output.strip() if len(output.strip()) > 0 else None
    except BaseException:
        # Didn't work, pass back None
        pass

    return user_name, user_email


def prepend_to_path_environment_variable(env, path_to_prepend):
    # On UNIX, it's PATH, on Windows it's Path
    variable_name = 'PATH'
    if 'Path' in env:
        variable_name = 'Path'

    path_variable = env[variable_name]

    path_variable = path_to_prepend + os.pathsep + path_variable

    env[path_variable] = path_variable

    return env


def get_md5_hash(file_name):
    import hashlib
    with open(file_name, 'rb') as file_to_check:
        data = file_to_check.read()
        return hashlib.md5(data).hexdigest()


def is_admin():
    import ctypes
    try:
        return os.getuid() == 0
    except AttributeError:
        return ctypes.windll.shell32.IsUserAnAdmin()


def fix_permissions(file_or_folder):
    if is_windows():
        # If this is Windows, we might be in an elevated process and that means we'll create the folder with
        # restricted access, which can cause problems later. Use icacls to give "Everyone" access. Administrators
        # can lock it down later.
        cmdline = 'icacls "%s" /grant *S-1-1-0:%sF >NUL' % (file_or_folder,
                                                            '(OI)(CI)' if os.path.isdir(file_or_folder) else '')
        spawn(cmdline)
    else:
        if os.path.isdir(file_or_folder):
            cmdline = f'chmod -R g=rwx,o=rwx "{file_or_folder}"'
        else:
            cmdline = f'chmod g=rw,o=rw "{file_or_folder}"'
        spawn(cmdline)


def create_folder_if_necessary_with_correct_permissions(folder):
    if not os.path.exists(folder):
        os.makedirs(folder)
        fix_permissions(folder)


def force_unicode(s):
    if isinstance(s, str):
        return s
    elif isinstance(s, bytes):
        return str(s, errors='replace')
    else:
        return str(s)


def force_print(s):
    if s is None:
        return

    print(force_unicode(s))


def get_branch_details():
    (ret_code, output, err) = spawn('git status', capture_output=True)

    if ret_code != 0:
        raise RuntimeError('Error calling "git status"')

    branch_name = None
    for match in re.finditer(r'On branch (.*?)\n', output):
        branch_name = match.group(1)

    if not branch_name:
        raise RuntimeError('ERROR: Could not discern git branch from "git status":\n%s' % output)

    pushed = False
    if 'Your branch is' in output:
        pushed = True
    else:
        # The local branch isn't tracking a remote one, but it may still have been pushed. Query directly.
        (ret_code_new, output_new, err_new) = spawn('git ls-remote origin ' + branch_name, capture_output=True)
        if branch_name in output_new:
            pushed = True

    matcher = re.match(r'.*crab-(\d+).*', branch_name.lower())
    if not matcher:
        return branch_name, None, pushed

    crab_number = matcher.group(1)

    return branch_name, crab_number, pushed


def human_readable_byte_count(bytes, use_base_two, round_to_int):
    import math
    """
    Python version of SystemInfo#humanReadableByteCount.
    Constructs a human readable string of the given bytes, such as '2.02GiB' or '9.75TB'.
    :param bytes: Number of bytes
    :param use_base_two: If true, use base-two units (GiB); if false, use base-ten (GB)
    :param round_to_int: If true, round the result to the nearest integer. If false, return the result rounded to two
    decimal points.
    :return: A human readable string of the given bytes, such as '2.02GiB' or '9.75TB'.
    """
    base_kilo = 1024 if use_base_two else 1000

    if bytes < base_kilo:
        return "%s B" % bytes

    exponent = int(math.log(bytes) // math.log(base_kilo))
    number = bytes / (base_kilo ** exponent)

    if round_to_int and (round(number) >= base_kilo):
        exponent += 1
        number = bytes / base_kilo ** exponent

    units = UNITS_BASE_TWO[exponent] if use_base_two else UNITS_BASE_TEN[exponent]

    if round_to_int:
        return "%d %s" % (round(number), units)
    else:
        return "%.2f %s" % (number, units)


def pytest(mark, cwd, parallelize=False, echo=False):
    parallel_arg = '-n auto' if parallelize else ''
    spawn(f'pytest -v -m {mark} --timeout 3600 {parallel_arg}', cwd=cwd, echo=echo)


def periodic_background_task(period: int, initial_delay=0, times=-1):
    """
    Annotate a function with e.g. `periodic_background_task(period=30, initial_delay=5)` to execute that function every
    30 seconds until the python script exits. The task is executed on a daemon thread created just for the task.

    :param period: Seconds between task executions as measured from the end of one execution to the start of the next.
    :param initial_delay: Seconds to delay the first task execution.
    :param times: Number of times to execute the task. -1 to execute forever.
    """

    def outer_wrap(function):
        def wrap(*args, **kwargs):
            stop = threading.Event()

            def inner_wrap():
                i = 0
                stop.wait(initial_delay)
                while i != times and not stop.isSet():
                    function(*args, **kwargs)
                    stop.wait(period)
                    i += 1

            t = threading.Timer(0, inner_wrap)
            t.daemon = True
            t.start()
            return stop

        return wrap

    return outer_wrap


def unpack(file_to_unpack, destination, is_targz=True):
    """ Unpack a .zip file or a .tar.gz file to a given destination
    :param file_to_unpack: String - a path to a .zip file or .tar.gz
    :param destination: String - a path to the destination directory
    :param is_targz: boolean - true if .tar.gz and false if .zip
    """
    import tarfile
    import zipfile
    print('Unpacking %s to %s' % (file_to_unpack, destination))
    if is_targz:
        t = tarfile.open(file_to_unpack, 'r:gz')
        t.extractall(destination)
        t.close()
    else:
        z = zipfile.ZipFile(file_to_unpack, "r")
        z.extractall(destination)
        z.close()
    os.remove(file_to_unpack)



import pandas as pd
import pytest

from seeq import spy
from seeq.sdk import *
from seeq.sdk.rest import ApiException
from seeq.spy import _login, Session
from seeq.spy import addons
from seeq.spy._errors import *
from seeq.spy.addons import _permissions
from seeq.spy.addons.tests import test_common as addons_test_common
from seeq.spy.tests import test_common
from seeq.spy.tests.test_common import Sessions


def setup_module():
    test_common.initialize_sessions()
    addons_test_common.enable_addon_tools(test_common.get_session(Sessions.admin), True)


def _create_testing_group():
    admin_session = test_common.get_session(Sessions.admin)
    user_groups_api = UserGroupsApi(admin_session.client)
    group_name = 'testers'

    # Note this code used to employ a read-then-write pattern where we would attempt to find the group and,
    # if not found, create it. This was vulnerable to a race condition when running tests in parallel: One test would
    # create the group in between the other test's find/create operations. Now we just try to create the group and if
    # that fails due to the group already existing, then find the group.

    group = None
    try:
        group = user_groups_api.create_user_group(body={"name": group_name})
    except ApiException as e:
        if e.status == 400:
            # This indicates that the group has already been created
            pass
        else:
            raise

    if group is None:
        group = _login.find_group(admin_session, group_name)

    return group


def _cleanup(name):
    admin_session = test_common.get_session(Sessions.admin)
    df = addons.search(pd.DataFrame([{'Name': f'{name}'}]), session=admin_session)
    if len(df) > 0:
        addons.uninstall(df, session=admin_session)


@pytest.mark.system
def test_addon_support():
    addons_test_common.test_addon_support(spy.session, 'install', dict())


def tool_names(session: Session):
    system_api = SystemApi(session.client)
    tools = system_api.get_add_on_tools().add_on_tools
    return [tool.name for tool in tools]


# noinspection HttpUrlsUsage
@pytest.mark.system
def test_install_one_tool():
    items_api = ItemsApi(spy.session.client)

    new_name = "test_install_one_tool"
    my_tool = {"Name": new_name,
               "Description": "My new tool",
               "Target URL": "http://www.google.com",
               "Groups": ["Everyone"]}

    try:
        with pytest.raises(Exception):
            # This will cause "Only administrators may create add-on tools" exception
            addons.install(my_tool)

        admin_session = test_common.get_session(Sessions.admin)
        df = addons.install(my_tool, session=admin_session)
        assert items_api.get_item_and_all_properties(id=df['ID'].values[0])
        permissions = _permissions.get_addon_permissions(df['ID'].values[0], items_api)
        assert 'Everyone' in permissions['Groups']
        assert new_name in tool_names(admin_session)

        with pytest.raises(Exception):
            # This will cause "Only administrators may delete add-on tools" exception
            addons.uninstall(df)

    finally:
        _cleanup(new_name)


# noinspection HttpUrlsUsage
@pytest.mark.system
def test_install_one_tool_with_existing_query_params():
    admin_session = test_common.get_session(Sessions.admin)
    items_api = ItemsApi(admin_session.client)

    new_name = "test_install_one_tool_with_existing_query_params"
    my_tool = {"Name": new_name,
               "Description": "My new tool",
               "Target URL": "http://www.google.com?previous=123",
               "Icon": "fa fa-icon"}

    try:
        df = addons.install(my_tool, include_workbook_parameters=True, session=admin_session)
        tool = items_api.get_item_and_all_properties(id=df['ID'].values[0])
        workbook_params = f'&workbookId={{workbookId}}&worksheetId={{worksheetId}}&workstepId={{' \
                          f'workstepId}}&seeqVersion={{seeqVersion}}'
        target_url = f"{my_tool['Target URL']}{workbook_params}"
        assert target_url == [x for x in tool.properties if x.name == 'Target URL'][0].value

    finally:
        _cleanup(new_name)


# noinspection HttpUrlsUsage
@pytest.mark.system
def test_install_one_tool_with_workbook_params():
    admin_session = test_common.get_session(Sessions.admin)
    items_api = ItemsApi(admin_session.client)

    new_name = f"test_install_one_tool_with_workbook_params"
    my_tool = {"Name": new_name,
               "Description": "My new tool",
               "Target URL": "http://www.google.com?",
               "Icon": "fa fa-icon"}

    try:
        df = addons.install(my_tool, include_workbook_parameters=True, session=admin_session)
        tool = items_api.get_item_and_all_properties(id=df['ID'].values[0])
        workbook_params = f'workbookId={{workbookId}}&worksheetId={{worksheetId}}&workstepId={{' \
                          f'workstepId}}&seeqVersion={{seeqVersion}}'
        target_url = f"{my_tool['Target URL']}{workbook_params}"
        assert target_url == [x for x in tool.properties if x.name == 'Target URL'][0].value

    finally:
        _cleanup(new_name)


# noinspection HttpUrlsUsage
@pytest.mark.system
def test_install_multiple_tools():
    admin_session = test_common.get_session(Sessions.admin)
    items_api = ItemsApi(admin_session.client)
    projects_api = ProjectsApi(admin_session.client)
    users_api = UsersApi(admin_session.client)

    user_id = users_api.get_me().id
    sdl_project1 = projects_api.create_project(body=dict(name='test project 1', ownerId=user_id))
    sdl_project2 = projects_api.create_project(body=dict(name='test project 1', ownerId=user_id))

    name = 'test_install_multiple_tools'
    new_tools = [
        {
            "Name": f'My {name}',
            "Description": "this is an awesome tool",
            "Target URL": f"http://www.my.seeq.com/data-lab/{sdl_project1.id}/apps/my_tool.ipynb",
            "Icon": "fa fa-bell"},
        {
            "Name": f'Your {name}',
            "Description": "this is an awesome tool",
            "Target URL": f"http://www.my.seeq.com/data-lab/{sdl_project2.id}/apps/tool2.ipynb",
            "Icon": "fa fa-bars"}]

    try:
        df = addons.install(new_tools, errors='raise', session=admin_session)
        for idd in df['ID'].values:
            assert items_api.get_item_and_all_properties(id=idd)

        for tool in new_tools:
            assert tool['Name'] in tool_names(admin_session)

    finally:
        _cleanup(f'*{name}')


# noinspection HttpUrlsUsage
@pytest.mark.system
def test_update_tool_false():
    admin_session = test_common.get_session(Sessions.admin)
    items_api = ItemsApi(admin_session.client)

    name = "test_update_tool_false"
    my_tool = {"Name": name,
               "Description": "My new tool",
               "Target URL": "http://www.google.com",
               "Icon": "fa fa-icon"}

    try:
        df = addons.install(my_tool, update_tool=False, session=admin_session)
        assert items_api.get_item_and_all_properties(id=df['ID'].values[0])
        assert name in tool_names(admin_session)
        with pytest.raises(SPyException,
                           match=f'Add-on tool with name "{name}" already exists. '
                                 f'You can update the existing tool with `update_tool=True`'):
            addons.install(my_tool, update_tool=False, session=admin_session)

    finally:
        _cleanup(name)


# noinspection HttpUrlsUsage
@pytest.mark.system
def test_update_tools_from_df():
    admin_session = test_common.get_session(Sessions.admin)
    projects_api = ProjectsApi(admin_session.client)
    users_api = UsersApi(admin_session.client)

    user_id = users_api.get_me().id
    sdl_project1 = projects_api.create_project(body=dict(name='test project 1', ownerId=user_id))
    sdl_project2 = projects_api.create_project(body=dict(name='test project 1', ownerId=user_id))

    name = 'test_update_tools_from_df'
    new_tools = [
        {
            "Name": f'My {name}',
            "Description": "this is an awesome tool",
            "Target URL": f"http://www.my.seeq.com/data-lab/{sdl_project1.id}/apps/my_tool.ipynb",
            "Icon": "fa fa-bell",
            "Link Type": "window"},
        {
            "Name": f'Your {name}',
            "Description": "this is an awesome tool",
            "Target URL": f"http://www.my.seeq.com/data-lab/{sdl_project2.id}/apps/tool2.ipynb",
            "Icon": "fa fa-bars",
            "Link Type": "window"}]

    try:
        df = addons.install(new_tools, session=admin_session)

        searched_tools = addons.search(df, session=admin_session)

        for _, tool in searched_tools.iterrows():
            assert tool['Link Type'] == 'window'

        searched_tools["Link Type"] = ['tab'] * len(searched_tools)
        addons.install(searched_tools, update_tool=True, session=admin_session)

        searched_tools = addons.search(df, session=admin_session)

        for _, tool in searched_tools.iterrows():
            assert tool['Link Type'] == 'tab'

    finally:
        _cleanup(f'*{name}')


# noinspection HttpUrlsUsage
@pytest.mark.system
def test_change_permissions():
    admin_session = test_common.get_session(Sessions.admin)
    projects_api = ProjectsApi(admin_session.client)
    users_api = UsersApi(admin_session.client)
    items_api = ItemsApi(admin_session.client)
    group = _create_testing_group()

    user_id = users_api.get_me().id
    sdl_project1 = projects_api.create_project(body=dict(name='test project 1', ownerId=user_id))
    sdl_project2 = projects_api.create_project(body=dict(name='test project 1', ownerId=user_id))
    name = 'test_change_permissions'
    new_tools = [
        {
            "Name": f'My {name}',
            "Description": "this is an awesome tool",
            "Target URL": f"http://www.my.seeq.com/data-lab/{sdl_project1.id}/apps/my_tool.ipynb",
            "Icon": "fa fa-bell",
            "Link Type": "window",
            "Groups": ["testers"]},
        {
            "Name": f'Your {name}',
            "Description": "this is an awesome tool",
            "Target URL": f"http://www.my.seeq.com/data-lab/{sdl_project2.id}/apps/tool2.ipynb",
            "Icon": "fa fa-bars",
            "Link Type": "window",
            "Groups": ["testers"]}]

    try:
        addons.install(new_tools, session=admin_session)

        search_df = addons.search({"Name": f'My test_change_permissions'}, session=admin_session)

        # Pass only the ID to make sure it's not grabbing other properties from the search_df
        idd = search_df['ID'][0]

        # Remove group permissions
        addons.install(pd.Series({"ID": idd, "Groups": []}), update_permissions=True, update_tool=False,
                       session=admin_session)

        permissions = _permissions.get_addon_permissions(idd, items_api)
        assert len(permissions['Groups']) == 0
        assert len(permissions["Users"]) == 0

        addons.install({"ID": idd, "Groups": [group.name]}, update_permissions=True, session=admin_session)

    finally:
        _cleanup(f'*{name}')


# noinspection HttpUrlsUsage
@pytest.mark.system
def test_update_tools_and_update_permissions_passing_id():
    _create_testing_group()

    projects_api = ProjectsApi(spy.session.client)
    users_api = UsersApi(spy.session.client)

    user_id = users_api.get_me().id
    sdl_project1 = projects_api.create_project(body=dict(name='test project 1', ownerId=user_id))
    sdl_project2 = projects_api.create_project(body=dict(name='test project 1', ownerId=user_id))

    name = 'test_update_tools_and_update_permissions_passing_id'
    new_tools = [
        {
            "Name": f'My {name}',
            "Description": "this is an awesome tool",
            "Target URL": f"http://www.my.seeq.com/data-lab/{sdl_project1.id}/apps/my_tool.ipynb",
            "Icon": "fa fa-bell",
            "Link Type": "window",
            "Groups": ["Everyone"]
        },
        {
            "Name": f'Your {name}',
            "Description": "this is an awesome tool",
            "Target URL": f"http://www.my.seeq.com/data-lab/{sdl_project2.id}/apps/tool2.ipynb",
            "Icon": "fa fa-bars",
            "Link Type": "window"}]

    admin_session = test_common.get_session(Sessions.admin)
    try:
        df = addons.install(new_tools, session=admin_session)

        searched_tools = addons.search(df)

        for _, tool in searched_tools.iterrows():
            assert tool['Link Type'] == 'window'

        searched_tools["Link Type"] = ['tab'] * len(searched_tools)
        searched_tools['Groups'] = [['testers']] * len(searched_tools)
        addons.install(searched_tools, update_tool=True, update_permissions=True, session=admin_session)

        searched_tools = addons.search(df)

        for _, tool in searched_tools.iterrows():
            assert tool['Link Type'] == 'tab'
            assert 'testers' in tool['Groups']
            assert 'Everyone' not in tool['Groups']

    finally:
        _cleanup(f'*{name}')


# noinspection HttpUrlsUsage
@pytest.mark.system
def test_mix_update_tool_and_new_tool():
    admin_session = test_common.get_session(Sessions.admin)
    items_api = ItemsApi(admin_session.client)

    name = "test_mix_update_tool_and_new_tool"
    my_tool = {"Name": name,
               "Description": "an awesome tool",
               "Target URL": "http://www.google.com?",
               "Icon": "fa fa-icon"}

    try:
        df = addons.install(my_tool, include_workbook_parameters=True, update_tool=True, session=admin_session)
        previous_id = df['ID'].values[0]
        tool = items_api.get_item_and_all_properties(id=previous_id)
        assert my_tool['Description'] == [x for x in tool.properties if x.name == 'Description'][0].value

        new_install = [{"Name": name,
                        "Description": "an awesome tool updated",
                        "Target URL": "http://www.google.com?",
                        "Icon": "fa fa-icon"},
                       {"Name": f'New {name}',
                        "Description": "this is new",
                        "Target URL": "http://www.google.com?",
                        "Icon": "fa fa-icon"}]

        df = addons.install(new_install, include_workbook_parameters=True, update_tool=True, session=admin_session)
        tool = items_api.get_item_and_all_properties(id=previous_id)
        tool2 = items_api.get_item_and_all_properties(id=df['ID'][1])
        assert new_install[0]['Description'] == [x for x in tool.properties if x.name == 'Description'][0].value
        assert new_install[1]['Description'] == [x for x in tool2.properties if x.name == 'Description'][0].value
        assert len(df) == 2
        assert df['Result'][0] == 'Updated'
        assert df['Result'][1] == 'Installed'

    finally:
        _cleanup(f'*{name}')


# noinspection HttpUrlsUsage
@pytest.mark.system
def test_update_duplicated_tool():
    admin_session = test_common.get_session(Sessions.admin)
    system_api = SystemApi(admin_session.client)
    name = "test_update_duplicated_tool"

    new_tool_config = dict(
        name=name,
        description="duplicated test",
        iconClass='fa fa-bars',
        targetUrl=f'https://www.seeq.com',
        linkType="window",
        windowDetails="toolbar=0,height=600,width=600",
        sortKey="a",
        # need to cast to bool if coming from pandas as np.bool_
        reuseWindow=True,
    )

    try:
        tool1_id = system_api.create_add_on_tool(body=new_tool_config).id
        system_api.create_add_on_tool(body=new_tool_config)

        my_tool = {"Name": name,
                   "Description": "an awesome tool",
                   "Target URL": "http://www.google.com?",
                   "Icon": "fa fa-icon"}

        # addons.install fails to update if update_tool=False
        with pytest.raises(SPyException,
                           match=f'Add-on tool with name "{name}" already exists. '
                                 f'You can update the existing tool with `update_tool=True`'):
            addons.install(my_tool, include_workbook_parameters=True, update_tool=False, session=admin_session)

        # addons.install fails to update even with update_tool=True since there are two tools with the same name
        with pytest.raises(SPyException,
                           match=f'There exists 2 tools with the name "{name}". '
                                 f'You can supply the ID to avoid ambiguity and modify only the intended tool'):
            addons.install(my_tool, include_workbook_parameters=True, update_tool=True, session=admin_session)

        my_tool = {"ID": tool1_id,
                   "Name": name,
                   "Description": "an awesome tool",
                   "Target URL": "http://www.google.com?",
                   "Icon": "fa fa-icon"}

        # passing the ID resolves the ambiguity and allows the update
        df = addons.install(my_tool, include_workbook_parameters=True, update_tool=True, session=admin_session)

        assert len(df) == 1
        assert df['Result'][0] == 'Updated'

    finally:
        _cleanup(name)



# coding: utf-8

"""
    Seeq REST API

    No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)

    OpenAPI spec version: 56.0.1-SNAPSHOT
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""


from pprint import pformat
from six import iteritems
import re


class PutScalarsInputV1(object):
    """
    NOTE: This class is auto generated by the swagger code generator program.
    Do not edit the class manually.
    """


    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'datasource_class': 'str',
        'datasource_id': 'str',
        'scalars': 'list[ScalarInputV1]'
    }

    attribute_map = {
        'datasource_class': 'datasourceClass',
        'datasource_id': 'datasourceId',
        'scalars': 'scalars'
    }

    def __init__(self, datasource_class=None, datasource_id=None, scalars=None):
        """
        PutScalarsInputV1 - a model defined in Swagger
        """

        self._datasource_class = None
        self._datasource_id = None
        self._scalars = None

        if datasource_class is not None:
          self.datasource_class = datasource_class
        if datasource_id is not None:
          self.datasource_id = datasource_id
        if scalars is not None:
          self.scalars = scalars

    @property
    def datasource_class(self):
        """
        Gets the datasource_class of this PutScalarsInputV1.
        Along with the Datasource ID, the Datasource Class uniquely identifies a datasource. For example, a datasource may be a particular instance of an OSIsoft PI historian.

        :return: The datasource_class of this PutScalarsInputV1.
        :rtype: str
        """
        return self._datasource_class

    @datasource_class.setter
    def datasource_class(self, datasource_class):
        """
        Sets the datasource_class of this PutScalarsInputV1.
        Along with the Datasource ID, the Datasource Class uniquely identifies a datasource. For example, a datasource may be a particular instance of an OSIsoft PI historian.

        :param datasource_class: The datasource_class of this PutScalarsInputV1.
        :type: str
        """
        if datasource_class is None:
            raise ValueError("Invalid value for `datasource_class`, must not be `None`")

        self._datasource_class = datasource_class

    @property
    def datasource_id(self):
        """
        Gets the datasource_id of this PutScalarsInputV1.
        Along with the Datasource Class, the Datasource ID uniquely identifies a datasource. For example, a datasource may be a particular instance of an OSIsoft PI historian.

        :return: The datasource_id of this PutScalarsInputV1.
        :rtype: str
        """
        return self._datasource_id

    @datasource_id.setter
    def datasource_id(self, datasource_id):
        """
        Sets the datasource_id of this PutScalarsInputV1.
        Along with the Datasource Class, the Datasource ID uniquely identifies a datasource. For example, a datasource may be a particular instance of an OSIsoft PI historian.

        :param datasource_id: The datasource_id of this PutScalarsInputV1.
        :type: str
        """
        if datasource_id is None:
            raise ValueError("Invalid value for `datasource_id`, must not be `None`")

        self._datasource_id = datasource_id

    @property
    def scalars(self):
        """
        Gets the scalars of this PutScalarsInputV1.
        A list of scalars to create or update

        :return: The scalars of this PutScalarsInputV1.
        :rtype: list[ScalarInputV1]
        """
        return self._scalars

    @scalars.setter
    def scalars(self, scalars):
        """
        Sets the scalars of this PutScalarsInputV1.
        A list of scalars to create or update

        :param scalars: The scalars of this PutScalarsInputV1.
        :type: list[ScalarInputV1]
        """
        if scalars is None:
            raise ValueError("Invalid value for `scalars`, must not be `None`")

        self._scalars = scalars

    def to_dict(self):
        """
        Returns the model properties as a dict
        """
        result = {}

        for attr, _ in iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """
        Returns the string representation of the model
        """
        return pformat(self.to_dict())

    def __repr__(self):
        """
        For `print` and `pprint`
        """
        return self.to_str()

    def __eq__(self, other):
        """
        Returns true if both objects are equal
        """
        if not isinstance(other, PutScalarsInputV1):
            return False

        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """
        Returns true if both objects are not equal
        """
        return not self == other
